"""These fixtures make it easy to mock out the loguru logger in tests.
"""
from unittest.mock import ANY

import pytest


@pytest.fixture
def mock_loguru_log(mocker):
    """Mock out the loguru logger for testing.

    Error works, not sure about the other levels. Check them out as you use them.
    """
    mock = mocker.patch("loguru.logger._log")
    mock.get_errors = lambda: get_loglines_at_level(mock, "error")
    mock.get_info = lambda: get_loglines_at_level(mock, "info")
    mock.get_warnings = lambda: get_loglines_at_level(mock, "warning")
    yield mock


def get_loglines_at_level(mock_loguru_log, level=ANY):
    return [call for call in mock_loguru_log.mock_calls if len(call.args) > 1 and str(call.args[0]).lower() == level]



"""Evolver uses duck-typing to give the appearance of editing a frozen, nested structure of attrs classes and tuples, recording changes in a way that they can be applied to generate a newly frozen instance.

One of the design goals is that mypy, autocomplete, and automatic refactoring work for the assignments made into these nested structures.

See: https://imbue-ai.slack.com/archives/C05D0SM2RT5/p1726185313480779?thread_ts=1722865932.537289&cid=C05D0SM2RT5

If you make changes here and then the tests fail with:
```
E   RecursionError: maximum recursion depth exceeded
!!! Recursion detected (same locals & position)
```
It's possible that you're accidentally invoking `Evolver.something_undefined` and that's causing the infinite recursion.
Mypy cannot catch this when it thinks the type is an `Evolver` because the `Evolver` class has a `__getattr__` method that makes it look like any attribute access could be valid.
"""
import threading
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import List
from typing import Tuple
from typing import Type
from typing import TypeGuard
from typing import TypeVar
from typing import Union
from typing import cast

import attr
from pydantic import BaseModel

from imbue_core.frozen_utils import FrozenDict
from imbue_core.pydantic_utils import model_update

_T = TypeVar("_T")
ObjectType = TypeVar("ObjectType")


def evolver(obj: _T) -> _T:
    """Creates a wrapper around an immutable attrs object, tuple, or FrozenDict that records potentially nested attribute assignments.

    The return type is our first white lie to the type system.
    """
    result = _Evolver[_T](obj)
    # The cast is a little white lie to the type system to make type-checking, autocomplete, and refactoring work.
    return cast(_T, result)


def assign(dest: _T, src: Callable[[], _T]) -> None:
    """Since mypy would complain about assignments to frozen attrs fields, use this function to make assignments.

    The only reason src is a `Callable[[], _T]` instead of just `_T` is that it makes type checking signal attempts to assign
    the wrong type to the field.  Surprisingly, just using `(dest: _T, src: _T)` doesn't cause mypy to complain about type mismatch.
    """
    assert isinstance(dest, _Evolver)  # Tricked you, type system!
    dest_evolver: _Evolver[_T] = cast(_Evolver[_T], dest)
    dest_evolver.assign(src())


def chill(evolver: _T) -> _T:
    """Produces a new frozen instance with the recorded changes applied.

    The name `chill` is a play on the fact that original input was frozen, and we are now re-freezing it.
    """
    assert isinstance(evolver, _Evolver)  # Tricked you, type system!
    cast_evolver = cast(_Evolver[_T], evolver)
    return cast_evolver.chill()


_threading_local = threading.local()


# TODO: since mutate and thaw are stateful, if you call one without the other, you run into problems.
def thaw(obj: _T) -> _T:
    global _threading_local
    if hasattr(_threading_local, "evolved_obj"):
        raise ValueError("Thaw does not support nested thawing.")
    _threading_local.evolved_obj = evolver(obj)
    return _threading_local.evolved_obj


# TODO: mypy complains because the input isn't anything related to ObjectType, but the output is.
# This also means the type checking doesn't quite work since it can't infer the return type of this function correctly
def mutate(dest: _T, src: Callable[[], _T]) -> ObjectType:  # type: ignore
    assign(dest, src)
    try:
        evolved_obj: ObjectType = _threading_local.evolved_obj
        return chill(evolved_obj)
    except AttributeError as e:
        raise ValueError("You must call mutate on a thawed object") from e
    finally:
        delattr(_threading_local, "evolved_obj")


def mutate_from_dict(dest: ObjectType, src: Dict[str, Any]) -> ObjectType:
    # Warning: using this function doesn't provide mypy type checking at the call site, but it allows a single interface for attrs and pydantic classes
    # In most cases the above function should be used instead
    evolved_obj = evolver(dest)
    for key, value in src.items():
        assign(getattr(evolved_obj, key), lambda: value)
    return chill(evolved_obj)


def evolver_isinstance(evolver: Any, cls: Type[_T]) -> TypeGuard[_T]:
    assert isinstance(evolver, _Evolver)  # Tricked you, type system!
    return evolver.isinstance(cls)


class _RegularValue:
    regular_value: Any

    def __init__(self, value: Any) -> None:
        self.regular_value = value


class _AttrValue:
    attr_value: Any
    child_evolver_by_name: dict[str, "_Evolver[Any]"]

    def __init__(self, value: Any) -> None:
        self.attr_value = value
        self.child_evolver_by_name = {}


class _PydanticModelValue:
    pydantic_model_value: Any
    child_evolver_by_name: dict[str, "_Evolver[Any]"]

    def __init__(self, value: Any) -> None:
        self.pydantic_model_value = value
        self.child_evolver_by_name = {}


class _TupleValue:
    tuple_evolvers: List["_Evolver[Any]"]

    def __init__(self, value: Tuple[Any, ...]) -> None:
        # It may be premature to create evolvers for all the elements of the tuple, but it's easier.
        self.tuple_evolvers = [evolver(item) for item in value]


class _FrozenDictValue:
    frozen_dict_evolvers: dict[Any, "_Evolver[Any]"]

    def __init__(self, value: Dict[Any, Any]) -> None:
        # It may be premature to create evolvers for all the elements of dict, but it's easier.
        self.frozen_dict_evolvers = {k: evolver(v) for k, v in value.items()}


class _Evolver(Generic[_T]):
    _value: Union[_RegularValue, _AttrValue, _TupleValue, _FrozenDictValue, _PydanticModelValue]

    def __init__(self, initial_value: _T) -> None:
        super().__init__()
        self.assign(initial_value)

    def assign(self, new_value: _T) -> None:
        """Assign a new value to this Evolver, recording a change to the frozen structure to be later applied during `chill()`."""

        if attr.has(type(new_value)):
            self._value = _AttrValue(new_value)
        elif isinstance(new_value, BaseModel):
            self._value = _PydanticModelValue(new_value)
        elif isinstance(new_value, tuple):
            self._value = _TupleValue(new_value)
        elif isinstance(new_value, FrozenDict):
            self._value = _FrozenDictValue(new_value)
        else:
            self._value = _RegularValue(new_value)

    def __getattr__(self, item: str) -> "_Evolver[Any]":
        """Access Evolvers for nested members of a frozen attrs object."""
        try:
            if isinstance(self._value, _AttrValue):
                if item not in self._value.child_evolver_by_name:
                    child_obj = getattr(self._value.attr_value, item)
                    result = evolver(child_obj)
                    assert isinstance(result, _Evolver), "Expose a lie to the type system."
                    self._value.child_evolver_by_name[item] = result
                return self._value.child_evolver_by_name[item]
            elif isinstance(self._value, _PydanticModelValue):
                if item not in self._value.child_evolver_by_name:
                    child_obj = getattr(self._value.pydantic_model_value, item)
                    result = evolver(child_obj)
                    assert isinstance(result, _Evolver), "Expose a lie to the type system."
                    self._value.child_evolver_by_name[item] = result
                return self._value.child_evolver_by_name[item]
            raise TypeError(
                f"You're trying to access field {item=} on an object of {type(self._value)=} that doesn't have that field (should have been a mypy error)."
            )
        except BaseException as e:
            if hasattr(_threading_local, "evolved_obj"):
                if getattr(_threading_local, "evolved_obj") == self:
                    delattr(_threading_local, "evolved_obj")
            raise e

    # TODO: It wouldn't be terribly difficult to support "appending" to the tuple as well, by appending to this list.
    def __getitem__(self, key: Any) -> "_Evolver[Any]":
        """Access Evolvers for the elements of a tuple or dict."""
        if isinstance(self._value, _TupleValue):
            assert isinstance(key, int)
            return self._value.tuple_evolvers[key]
        elif isinstance(self._value, _FrozenDictValue):
            if key not in self._value.frozen_dict_evolvers:
                # Presumably we're going to evolver_assign to this very soon.
                self._value.frozen_dict_evolvers[key] = _Evolver(_RegularValue(None))
            return self._value.frozen_dict_evolvers[key]
        raise TypeError(
            f"You're using [square_brackets] access {key=} on an object of {type(self._value)=} that doesn't support this (should have been a mypy error)."
        )

    def chill(self) -> _T:
        """Recursively apply the recorded changes to the original object and return a new frozen instance."""
        if isinstance(self._value, _AttrValue):
            new_children: dict[str, Any] = {
                name: chill(child) for name, child in self._value.child_evolver_by_name.items()
            }
            assert attr.has(self._value.attr_value.__class__)
            return cast(_T, attr.evolve(cast(Any, self._value.attr_value), **new_children))
        elif isinstance(self._value, _PydanticModelValue):
            return cast(
                _T,
                model_update(
                    self._value.pydantic_model_value,
                    update={name: chill(child) for name, child in self._value.child_evolver_by_name.items()},
                ),
            )
        elif isinstance(self._value, _TupleValue):
            return cast(_T, tuple(evolver.chill() for evolver in self._value.tuple_evolvers))
        elif isinstance(self._value, _RegularValue):
            return cast(_T, self._value.regular_value)
        elif isinstance(self._value, _FrozenDictValue):
            return cast(_T, FrozenDict({k: v.chill() for k, v in self._value.frozen_dict_evolvers.items()}))
        raise ValueError(f"This Evolver has no value to evolve, {type(self._value)=}.")

    def isinstance(self, cls: Type[_T]) -> bool:
        """Check if the object being evolved is an instance of the given class."""
        if isinstance(self._value, _AttrValue):
            return isinstance(self._value.attr_value, cls)
        elif isinstance(self._value, _PydanticModelValue):
            return isinstance(self._value.pydantic_model_value, cls)
        elif isinstance(self._value, _FrozenDictValue):
            return cls == FrozenDict
        return False



import asyncio
import functools
import hashlib
from pathlib import Path
from typing import Mapping
from typing import Tuple

import anyio
import attr

from imbue_core.cattrs_serialization import serialize_to_json
from imbue_core.computing_environment.computing_environment import apply_patch_via_git
from imbue_core.computing_environment.computing_environment import make_commit
from imbue_core.frozen_utils import empty_mapping
from imbue_core.git import LocalGitRepo
from imbue_core.secrets_utils import get_secret


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class GitRepo:
    git_user_name: str
    git_user_email: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class GitRepoSnapshot(GitRepo):
    git_hash: str
    git_branch: str
    git_diff: str | None

    @functools.cached_property
    def reference_hash(self) -> str:
        hash_fn = hashlib.md5()
        hash_fn.update(serialize_to_json(self).encode("UTF-8"))
        return hash_fn.hexdigest()


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class RemoteGitRepoSnapshot(GitRepoSnapshot):
    git_repo_url: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class LocalGitRepoSnapshot(GitRepoSnapshot):
    git_repo_path: str
    # the relative paths of any untracked files in the repo
    git_untracked_files: Tuple[str, ...] | None


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class GitCommitSnapshot:
    contents_by_path: Mapping[str, str]
    commit_message: str
    # ex: "2023-05-15T14:30:00"
    # used for GIT_AUTHOR_DATE and GIT_COMMITTER_DATE
    commit_time: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class FullLocalGitRepo(GitRepo):
    main_history: Tuple[GitCommitSnapshot, ...]
    # the relative paths of any currently untracked files in the repo, and their content
    untracked_file_content_by_path: Mapping[str, str] = attr.ib(factory=empty_mapping)
    git_branch: str = "main"
    git_diff: str | None = None


async def create_repo_from_snapshot(
    full_repo: FullLocalGitRepo,
    destination_path: Path,
) -> LocalGitRepo:
    """Creates an entire repo history locally from scratch. Much faster and more reliable than checking from remote."""
    anyio_path = anyio.Path(destination_path)
    if not await anyio_path.exists():
        destination_path.mkdir(parents=True)
    assert (
        await anyio_path.is_dir()
    ), f"Destination for repo checkout must be a directory. {destination_path} is not a directory."

    async for _ in anyio_path.iterdir():
        raise Exception(f"Destination for repo creation must be an empty directory. {destination_path} is not empty.")

    # create the empty repo
    new_repo = LocalGitRepo(destination_path)
    await new_repo.run_git(("init",))
    await new_repo.run_git(("checkout", "-b", "main"))
    if full_repo.git_branch != "main":
        await new_repo.run_git(("branch", "-m", "main", full_repo.git_branch))
    await new_repo.run_git(("config", "user.name", f"'{full_repo.git_user_name}'"))
    await new_repo.run_git(("config", "user.email", f"'{full_repo.git_user_email}'"))

    # put the history in
    for commit in full_repo.main_history:
        await _write_files_in_parallel(new_repo, commit.contents_by_path)
        await make_commit(new_repo, commit.commit_message, commit_time=commit.commit_time)

    if full_repo.git_diff:
        # apply any diffs from between git_hash and repo snapshot state of repo being checked out
        await apply_patch_via_git(new_repo, full_repo.git_diff, is_error_logged=True)

    if full_repo.untracked_file_content_by_path:
        # make sure the untracked file contents are there
        await _write_files_in_parallel(new_repo, full_repo.untracked_file_content_by_path)

    return new_repo


async def checkout_repo_from_snapshot(
    repo_snapshot: LocalGitRepoSnapshot | RemoteGitRepoSnapshot,
    destination_path: Path,
) -> LocalGitRepo:
    """Checks out an existing repo into a directory in the fastest way possible.

    Uses a snapshot of the existing repo to checkout a specific commit and branch, and apply any uncommited changes.
    Will checkout from a remote or local repo depending on the type of snapshot.

    Note, checking out from a remote repo requires that the necessary git permissions, etc are configured.

    See here: https://stackoverflow.com/questions/31278902/how-to-shallow-clone-a-specific-commit-with-depth-1
    """
    anyio_path = anyio.Path(destination_path)
    assert (
        await anyio_path.is_dir()
    ), f"Destination for repo checkout must be a directory. {destination_path} is not a directory."
    if not await anyio_path.exists():
        destination_path.mkdir(parents=True)
    async for _ in anyio_path.iterdir():
        raise Exception(f"Destination for repo checkout must be an empty directory. {destination_path} is not empty.")

    remote_address: str | None = None
    if isinstance(repo_snapshot, RemoteGitRepoSnapshot):
        token = get_secret("GIT_TOKEN")
        assert token is not None, "Must set GIT_TOKEN environment variable to clone git repos"
        env = {"GIT_TOKEN": token}

        repo_url = repo_snapshot.git_repo_url
        assert repo_url.startswith("https://"), "Only https git urls are supported"
        if repo_url.startswith("https://oauth2"):
            raise Exception("Wait no, that doesn't make sense--that will hardcode the oauth token into the DB")
        remote_address = repo_url.replace("https://", f"https://oauth2:{token}@", 1)
    elif isinstance(repo_snapshot, LocalGitRepoSnapshot):
        remote_address = repo_snapshot.git_repo_path
        env = None
    assert remote_address is not None, "Remote address not found"

    new_repo = LocalGitRepo(destination_path)
    await new_repo.run_git(("init",))
    await new_repo.run_git(("config", "user.name", f"'{repo_snapshot.git_user_name}'"))
    await new_repo.run_git(("config", "user.email", f"'{repo_snapshot.git_user_email}'"))
    await new_repo.run_git(("remote", "add", "origin", str(remote_address)))
    await new_repo.run_command(("git", "fetch", "--depth", "1", "origin", repo_snapshot.git_hash), secrets=env)
    await new_repo.run_git(("checkout", "FETCH_HEAD"))

    if repo_snapshot.git_diff:
        # apply any diffs from between git_hash and repo snapshot state of repo being checked out
        await apply_patch_via_git(new_repo, repo_snapshot.git_diff, is_error_logged=True)

    return new_repo


async def _write_files_in_parallel(repo: LocalGitRepo, content: Mapping[str, str]) -> None:
    async with asyncio.TaskGroup() as tg:
        tasks = [tg.create_task(repo.write_file(file_path, content)) for file_path, content in content.items()]
        await asyncio.gather(*tasks)


async def get_snapshot_info(repo: LocalGitRepo) -> GitRepoSnapshot:
    """Get general snapshot of the current state of the git repo."""
    async with asyncio.TaskGroup() as tg:
        git_branch_task = tg.create_task(repo.run_git(("rev-parse", "--abbrev-ref", "HEAD")))
        git_hash_task = tg.create_task(repo.run_git(("rev-parse", "HEAD")))
        git_unstaged_diff_task = tg.create_task(repo.run_git(("diff", "--full-index", "--binary"), is_stripped=False))
        git_staged_diff_task = tg.create_task(
            repo.run_git(("diff", "--full-index", "--binary", "--staged"), is_stripped=False)
        )
        git_user_name_task = tg.create_task(repo.run_git(("config", "user.name")))
        git_user_email_task = tg.create_task(repo.run_git(("config", "user.email")))
        await asyncio.gather(
            *[
                git_branch_task,
                git_hash_task,
                git_unstaged_diff_task,
                git_staged_diff_task,
                git_user_name_task,
                git_user_email_task,
            ]
        )
        git_branch = git_branch_task.result()
        current_git_hash = git_hash_task.result()
        # get the current diff (changes the user has made)
        git_staged_diff = git_staged_diff_task.result()
        git_diff = git_staged_diff
        git_unstaged_diff = git_unstaged_diff_task.result()
        if git_unstaged_diff.strip() != "":
            git_diff += git_unstaged_diff

        git_user_email = git_user_email_task.result()
        git_user_name = git_user_name_task.result()

        return GitRepoSnapshot(
            git_hash=current_git_hash,
            git_diff=git_diff,
            git_branch=git_branch,
            git_user_name=git_user_name,
            git_user_email=git_user_email,
        )


async def get_local_repo_snapshot(repo: LocalGitRepo) -> LocalGitRepoSnapshot:
    """Get a snapshot of the current state of the git repo locally."""
    # run a bunch of commands in parallel to generate the necessary information
    async with asyncio.TaskGroup() as tg:
        general_snapshot_task = tg.create_task(get_snapshot_info(repo))
        git_untracked_files_task = tg.create_task(repo.run_git(("ls-files", "--others", "--exclude-standard")))
        await asyncio.gather(
            *[
                general_snapshot_task,
                git_untracked_files_task,
            ]
        )
        # relative path to any untracked files in repo (that are not in excluded files, etc)
        untracked_files_result = git_untracked_files_task.result()
        if untracked_files_result not in (None, ""):
            untracked_files = tuple(untracked_files_result.splitlines())
        else:
            untracked_files = None

        general_snapshot = general_snapshot_task.result()
    return LocalGitRepoSnapshot(
        git_repo_path=str(repo.base_path),
        git_hash=general_snapshot.git_hash,
        git_diff=general_snapshot.git_diff,
        git_branch=general_snapshot.git_branch,
        git_untracked_files=untracked_files,
        git_user_email=general_snapshot.git_user_email,
        git_user_name=general_snapshot.git_user_name,
    )


async def get_repo_snapshot_for_cwd() -> GitRepoSnapshot:
    return await get_local_repo_snapshot(LocalGitRepo.build_from_cwd())



from typing import Any
from typing import Dict
from typing import TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


def model_update(model: T, update: Dict[str, Any]) -> T:
    """
    Update a Pydantic model with a dictionary of updates.
    Validation is performed to ensure items in the update dictionary are valid fields in the model.
    Use the Evolver class (imbue_core/imbue_core/nested_evolver.py) for type checking

    Args:
        model (BaseModel): The original Pydantic model.
        update (Dict[str, Any]): A dictionary of updates to apply to the model.

    Returns:
        BaseModel: A new instance of the model with the updates applied.

    """
    update_dict_fields = update.keys()
    model_fields = set(model.__class__.model_fields)
    extra_fields = update_dict_fields - model_fields
    if extra_fields:
        raise ValueError(f"Invalid fields: {extra_fields}")
    return fields_only_model_copy(model, update=update)


def fields_only_model_copy(model: T, update: Dict[str, Any] = {}) -> T:
    """
    Create a copy of a Pydantic model with only the fields defined in the model.

    (Specifically, do not copy cached properties.)

    """
    fields = {name: update.get(name, getattr(model, name)) for name in model.__class__.model_fields}
    return model.__class__(**fields)



import asyncio
import contextlib
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile
from tempfile import TemporaryDirectory
from typing import Generator

from imbue_core.common import TEMP_DIR
from imbue_core.git import LocalGitRepo
from imbue_core.git import get_git_repo_root
from imbue_core.testing_utils import fixture


@contextlib.contextmanager
def create_temp_file(contents: str, suffix: str, root_dir: Path) -> Generator[Path, None, None]:
    with NamedTemporaryFile(mode="w", suffix=suffix, dir=root_dir, delete=False) as temp_file:
        temp_file.write(contents)
        temp_file.flush()
        yield Path(temp_file.name)
        temp_file.close()
        Path(temp_file.name).unlink()


@contextlib.contextmanager
def create_temp_dir(root_dir: Path) -> Generator[Path, None, None]:
    with TemporaryDirectory(dir=root_dir) as temp_dir:
        yield Path(temp_dir)
        shutil.rmtree(temp_dir)


async def _configure_temp_mock_repo(mock_repo: LocalGitRepo, temp_dir: Path) -> LocalGitRepo:
    temp_repo = await mock_repo.copy_repo(temp_dir)
    await temp_repo.configure_git(
        git_user_name="AGI (Automated Software Inspector)",
        git_user_email="the_true_AGI@running.pytest.com",
    )
    return temp_repo


@fixture
def mock_repo_() -> Generator[LocalGitRepo, None, None]:
    mock_repo_path = get_git_repo_root() / "imbue/imbue/test_data/mock_repo"
    mock_repo = LocalGitRepo(base_path=mock_repo_path)
    with create_temp_dir(root_dir=Path(TEMP_DIR)) as temp_dir:
        temp_repo = asyncio.run(_configure_temp_mock_repo(mock_repo, temp_dir))
        yield temp_repo



import hashlib
import inspect
import os
import platform
import uuid
from pathlib import Path
from types import FrameType
from typing import Final
from typing import List

import pathspec


def is_on_osx() -> bool:
    return platform.system().lower() == "darwin"


def is_testing() -> bool:
    return "PYTEST_CURRENT_TEST" in os.environ


def _get_filesystem_root() -> str:
    env_value = os.getenv("SCIENCE_FILESYSTEM_ROOT")
    if not env_value:
        if is_on_osx():
            return "/tmp/science"
        else:
            # When on the physical cluster (and possibly other core clusters), this path is mounted to a unique per-container file path.
            # Anything produced at runtime >10mb should likely go here, as well as anything you might want to dig up for later debugging.
            # The hosts clean up the paths from dead containers periodically, but large data processing jobs should still clean up after themselves.
            return "/mnt/private"
    return env_value


FILESYSTEM_ROOT: Final = _get_filesystem_root()
TEMP_DIR = os.path.join(FILESYSTEM_ROOT, "tmp")

if not is_on_osx():
    os.makedirs(TEMP_DIR, exist_ok=True)


def hash_string(string: str) -> str:
    return hashlib.md5(string.encode("utf-8")).hexdigest()


def get_current_function_name() -> str:
    frame = inspect.currentframe()
    if frame is None:
        return "no_frame"
    prev_frame = frame.f_back
    if prev_frame is None or not isinstance(prev_frame, FrameType):
        return "no_previous_frame"
    return prev_frame.f_code.co_name


def filter_excluded_files(files: List[Path], directory: Path, exclude_file_name: str = ".gitignore") -> List[Path]:
    """Remove files from the list that are matched by a .gitignore or similarly-specified exclude file such as
    excluded.txt"""

    # Underneath the root directory, find all the excluders.
    # They can occur in subfolders and if they do they apply only to that subfolder.
    excluders = {path for path in directory.rglob(exclude_file_name) if not path.is_symlink()}

    # Per excluder, make a pathspec.
    for excluder in excluders:
        with excluder.open("r") as exclude_file:
            exclude_spec = pathspec.GitIgnoreSpec.from_lines(exclude_file)

            # Now we have two cases - We keep the file if the excluder doesn't apply because it's in a different
            # folder, or if it applies but doesn't match
            prefix = os.path.dirname(excluder)
            files = [
                file
                for file in files
                if not (file.is_relative_to(prefix) and exclude_spec.match_file(file.relative_to(prefix)))
            ]

    return files


def generate_id() -> str:
    return uuid.uuid4().hex


def generate_id_from_existing_id(existing_id: str, seed: int) -> str:
    return hashlib.md5(f"{existing_id}-{seed}".encode()).hexdigest()


def truncate_string(s: str, max_length: int) -> str:
    if len(s) <= max_length:
        return s
    return s[: max_length - 3] + "..."



from __future__ import annotations

import asyncio
import shlex
import time
from datetime import datetime
from pathlib import Path
from typing import Dict
from typing import Optional
from typing import Protocol
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union
from uuid import uuid4

import anyio
from loguru import logger
from tenacity import TryAgain
from tenacity import retry
from tenacity import retry_all
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_random_exponential

from imbue_core.async_monkey_patches import log_exception
from imbue_core.computing_environment.data_types import AnyPath
from imbue_core.computing_environment.data_types import FailedToMakeCommitError
from imbue_core.computing_environment.data_types import PatchApplicationError
from imbue_core.computing_environment.data_types import RunCommandError
from imbue_core.git_data_types import CommitTimestamp
from imbue_core.retry_utils import log_before_sleep
from imbue_core.section import Section
from imbue_core.time_utils import get_current_time

# Import the types needed for file modes
if TYPE_CHECKING:
    # for proper file mode typing
    from _typeshed import OpenBinaryModeReading
    from _typeshed import OpenBinaryModeWriting
    from _typeshed import OpenTextModeReading
    from _typeshed import OpenTextModeWriting


class ComputingEnvironment(Protocol):
    """Protocol defining the interface for a computing environment.

    This protocol specifies the required methods for interacting with a computing
    environment, including running commands and file operations.
    """

    async def run_command(
        self,
        command: Sequence[str],
        check: bool = True,
        secrets: Optional[Dict[str, str]] = None,
        cwd: Optional[AnyPath] = None,
        is_error_logged: bool = True,
    ) -> str:
        ...

    async def run_git(
        self,
        command: Sequence[str],
        check: bool = True,
        cwd: Optional[AnyPath] = None,
        is_error_logged: bool = True,
        is_stripped: bool = True,
        retry_on_git_lock_error: bool = True,
    ) -> str:
        ...

    async def write_file(
        self,
        relative_path: AnyPath,
        content: Optional[Union[str, bytes]],
        cwd: Optional[AnyPath] = None,
        mode: Union[OpenTextModeWriting, OpenBinaryModeWriting] = "w",
        mkdir_if_missing: bool = True,
    ) -> None:
        ...

    async def read_file(
        self,
        relative_path: AnyPath,
        cwd: Optional[AnyPath] = None,
        mode: Union[OpenTextModeReading, OpenBinaryModeReading] = "r",
        mkdir_if_missing: bool = True,
    ) -> Union[str, bytes]:
        ...

    async def delete_file(
        self,
        relative_path: AnyPath,
        cwd: Optional[AnyPath] = None,
    ) -> None:
        ...


def _get_temp_patch_file() -> anyio.Path:
    # this is a bad idea because it triggers the file watcher
    # patch_file = (self.base_path / str(uuid4())).with_suffix(".patch")
    patch_file = (Path("/tmp") / uuid4().hex).with_suffix(".patch")
    return anyio.Path(patch_file)


async def run_command_with_retry_on_git_lock_error(
    computing_environment: ComputingEnvironment,
    command: Sequence[str],
    check: bool = True,
    is_error_logged: bool = True,
    cwd: Optional[AnyPath] = None,
) -> str:
    max_retries = 50
    retry_count = 0
    retry_delay = 0.1  # seconds
    while True:
        try:
            return await computing_environment.run_command(
                command, check=check, is_error_logged=is_error_logged and retry_count >= max_retries, cwd=cwd
            )
        except RunCommandError as e:
            error_message = str(e)
            if "fatal: Unable to create" in error_message and ".git/index.lock': File exists" in error_message:
                if retry_count >= max_retries:
                    raise
                await asyncio.sleep(retry_delay)
                retry_count += 1
            else:
                raise


@retry(
    wait=wait_random_exponential(multiplier=0.1, max=2, exp_base=2),
    reraise=True,
    stop=stop_after_attempt(50),
    before_sleep=log_before_sleep,
)
async def wait_for_git_index_lock_to_be_free(local_sync_repo_path: Path) -> None:
    # Path to the git index lock file using anyio.Path to avoid blocking
    lock_file_path = anyio.Path(local_sync_repo_path / ".git" / "index.lock")
    if await lock_file_path.exists():
        raise TryAgain


async def apply_patch_without_git(computing_environment: ComputingEnvironment, diff: str) -> None:
    if diff.strip() == "":
        return
    patch_file = _get_temp_patch_file()
    try:
        await computing_environment.write_file(patch_file, diff)
        await computing_environment.run_command(("bash", "-c", f"patch -p1 < {patch_file}"))
    except RunCommandError as e:
        raise PatchApplicationError(f"Failed to apply patch: {e}") from e
    finally:
        await computing_environment.delete_file(patch_file)


async def is_repo_dirty(computing_environment: ComputingEnvironment, is_untracked_considered: bool = True) -> bool:
    """Check if the repo has any uncommitted changes."""
    return bool(
        await computing_environment.run_git(
            ("status", "--porcelain", *([] if is_untracked_considered else ["--untracked-files=no"]))
        )
    )


async def are_all_commits_pushed(computing_environment: ComputingEnvironment) -> bool:
    """Check if the repo has any unpushed commits."""
    output = await computing_environment.run_git(("cherry",))
    return output.strip() == ""


async def are_all_remote_commits_pulled(computing_environment: ComputingEnvironment) -> bool:
    # note this will fail if the branch hasn't been pushed to the remote
    output = await computing_environment.run_command(
        ("bash", "-c", "git fetch && git rev-list HEAD..@{upstream} --count")
    )
    return output.strip() == "0"


async def assert_repo_is_clean(computing_environment: ComputingEnvironment) -> None:
    """Assert that the repo has no uncommitted changes."""
    assert not await is_repo_dirty(
        computing_environment
    ), "You have untracked files. Please address them before using this script (this is to prevent accidentally adding large files unintentionally)"


async def get_branch_name(computing_environment: ComputingEnvironment, is_error_logged: bool = True) -> str:
    """Get the name of the current branch."""
    return await computing_environment.run_git(("symbolic-ref", "--short", "HEAD"), is_error_logged=is_error_logged)


async def rename_branch(
    computing_environment: ComputingEnvironment, old_name: str, new_name: str, force_if_exists: bool = True
) -> None:
    """Rename the given branch."""
    if force_if_exists:
        await computing_environment.run_git(("branch", "-M", old_name, new_name))
    else:
        await computing_environment.run_git(("branch", "-m", old_name, new_name))


async def get_branch_description(computing_environment: ComputingEnvironment, branch_name: str) -> str:
    """Get the description of the given branch."""
    try:
        return await computing_environment.run_git(
            ("config", f"branch.{branch_name}.description"), is_error_logged=False
        )
    except RunCommandError as e:
        if e.returncode == 1:
            # no description set
            return ""
        raise


async def is_branch_exists(computing_environment: ComputingEnvironment, branch_name: str) -> bool:
    """Check if the given branch exists."""
    result = await computing_environment.run_git(
        ("rev-parse", "--verify", "--quiet", branch_name), is_error_logged=False, check=False
    )
    return result.strip() != ""


async def is_commit_a_branch(computing_environment: ComputingEnvironment, commit_hash: str) -> bool:
    """Check if the given commit is a branch."""
    try:
        await computing_environment.run_git(
            ("show-ref", "--verify", "-q", f"refs/heads/{commit_hash}"), is_error_logged=False, check=True
        )
        return True
    except RunCommandError as e:
        if e.returncode == 1:
            return False
        raise


async def is_detached_head(computing_environment: ComputingEnvironment) -> bool:
    """Check if the current HEAD is detached."""
    result = await computing_environment.run_git(("rev-parse", "--abbrev-ref", "HEAD"), is_error_logged=False)
    return result.strip() == "HEAD"


async def set_branch_description(
    computing_environment: ComputingEnvironment, branch_name: str, description: str
) -> None:
    """Set the description of the given branch."""
    await computing_environment.run_git(("config", f"branch.{branch_name}.description", description))


async def get_branch_commit(computing_environment: ComputingEnvironment, branch_name: str) -> str:
    """Get the commit of the given branch."""
    return await computing_environment.run_git(("rev-parse", branch_name))


async def get_all_branch_names_pointing_to_commit(
    computing_environment: ComputingEnvironment, commit_hash: str
) -> Tuple[str, ...]:
    """Get all branch names that point to the given commit."""
    result = await computing_environment.run_git(
        ("for-each-ref", "refs/heads/", "--format='%(refname:short)'", "--points-at", commit_hash)
    )
    branch_names = tuple(result.splitlines())
    # strip the quotes
    return tuple(branch_name.strip("'") for branch_name in branch_names)


async def is_branch_child_of_branch(
    computing_environment: ComputingEnvironment, child_branch_name: str, parent_branch_name: str
) -> bool:
    """Check if the given branch is a child of the parent branch."""
    try:
        await computing_environment.run_git(
            ("merge-base", "--is-ancestor", parent_branch_name, child_branch_name), is_error_logged=False
        )
        return True
    except RunCommandError as e:
        if e.stderr.strip() == "" and e.returncode == 1:
            # we expect this command to give an empty stderr and a return code of 1
            # if the child branch is not an ancestor of the parent branch
            return False
        raise


async def is_commit_on_branch(
    computing_environment: ComputingEnvironment, commit_hash: str, branch_name: str, local_only: bool = True
) -> bool:
    """Check if the given commit is on the given branch."""
    if local_only:
        result = await computing_environment.run_git(("branch", "--contains", commit_hash))
    else:
        result = await computing_environment.run_git(("branch", "-a", "--contains", commit_hash))
    return any(branch_name == x.strip() for x in result.splitlines())


async def fetch_and_get_first_entry_in_fetch_head(
    computing_environment: ComputingEnvironment, remote: str, fetch_refs: Sequence[str]
) -> str:
    """Fetch the given refs from the remote and return the first entry in FETCH_HEAD."""
    refs_str = " ".join(fetch_refs)
    command = [
        "bash",
        "-c",
        (
            f"git fetch {remote} {refs_str} && "
            # get first commit from FETCH_HEAD
            "git rev-parse FETCH_HEAD"
        ),
    ]
    result = await run_command_with_retry_on_git_lock_error(computing_environment, command)
    return result.strip()


async def fetch_branch(computing_environment: ComputingEnvironment, branch_name: str) -> None:
    """Fetch the given branch from the remote."""
    await computing_environment.run_git(("fetch", "origin", branch_name))


async def is_branch_present(computing_environment: ComputingEnvironment, branch_name: str) -> bool:
    """Check if branch with given name is present."""
    result = await computing_environment.run_git(("branch",))
    return branch_name in result.splitlines()


async def create_reset_and_checkout_branch(computing_environment: ComputingEnvironment, branch_name: str) -> str:
    """Create new branch with given name."""
    return await computing_environment.run_git(("switch", "-C", branch_name))


async def switch_branch(computing_environment: ComputingEnvironment, branch_name: str) -> str:
    """Switch to branch with given name."""
    return await computing_environment.run_git(("switch", branch_name))


async def delete_branch(computing_environment: ComputingEnvironment, branch_name: str, delete_remote: bool) -> str:
    """Delete branch with given name."""
    result = await computing_environment.run_git(("branch", "-D", branch_name))
    if delete_remote:
        result = await computing_environment.run_git(("push", "origin", "--delete", branch_name))
    return result


async def update_branch_to_hash(computing_environment: ComputingEnvironment, branch_name: str, git_hash: str) -> None:
    """Update the given branch to reference the given git hash."""
    # here we do it without checking out the branch
    await computing_environment.run_git(("branch", "-f", branch_name, git_hash))


async def switch_and_create_branch_if_needed(computing_environment: ComputingEnvironment, branch_name: str) -> str:
    """Switch to new branch, creating it if it doesn't already exist."""
    if await is_branch_present(computing_environment, branch_name):
        await switch_branch(computing_environment, branch_name)
    else:
        await create_reset_and_checkout_branch(computing_environment, branch_name)
    return await get_branch_name(computing_environment)


async def merge_branches(
    computing_environment: ComputingEnvironment,
    base_branch_name: str,
    merge_branch_name: str,
    is_moving_to_base_branch: bool = True,
) -> str:
    """Merge `merge_branch_name` into `base_branch_name`."""
    await switch_branch(computing_environment, base_branch_name)
    await computing_environment.run_git(("merge", merge_branch_name))
    if not is_moving_to_base_branch:
        await switch_branch(computing_environment, "-")
    return await get_branch_name(computing_environment)


async def get_merge_base(computing_environment: ComputingEnvironment, branch_name: str, target_branch: str) -> str:
    """Get the merge base of the given branch and target branch.

    The merge base is the most recent commit that is on both branches.
    """
    return await computing_environment.run_git(["merge-base", branch_name, target_branch], is_error_logged=False)


async def checkout_hash(computing_environment: ComputingEnvironment, git_hash: str) -> str:
    """Checkout given git hash."""
    return await computing_environment.run_git(("checkout", git_hash))


async def force_add(computing_environment: ComputingEnvironment, *paths: str) -> None:
    """Force-add the specified paths to the git index."""
    await computing_environment.run_git(("add", "-f", *paths))


async def git_add(computing_environment: ComputingEnvironment, *paths: str) -> None:
    """Add the specified paths to the git index."""
    await computing_environment.run_git(("add", *paths))


def convert_datetime_to_git_timestamp(dt: datetime) -> str:
    return datetime.isoformat(dt)


def convert_git_timestamp_to_datetime(timestamp: str) -> datetime:
    return datetime.fromisoformat(timestamp)


def get_commit_ts_for_current_time() -> CommitTimestamp:
    """Get the commit timestamp for the current time."""
    current_time = get_current_time()
    return CommitTimestamp(
        author_ts=convert_datetime_to_git_timestamp(current_time),
        committer_ts=convert_datetime_to_git_timestamp(current_time),
    )


def _convert_time_to_commit_ts(time: str | datetime | CommitTimestamp | None) -> CommitTimestamp:
    if time is None:
        return get_commit_ts_for_current_time()
    elif isinstance(time, datetime):
        return CommitTimestamp(
            author_ts=convert_datetime_to_git_timestamp(time), committer_ts=convert_datetime_to_git_timestamp(time)
        )
    elif isinstance(time, CommitTimestamp):
        return time
    else:
        # assume it's a git timestamp
        return CommitTimestamp(author_ts=time, committer_ts=time)


async def make_commit(
    computing_environment: ComputingEnvironment,
    commit_message: str,
    allow_empty: bool = False,
    amend: bool = False,
    commit_time: str | datetime | CommitTimestamp | None = None,
) -> str:
    if commit_message.strip() == "":
        commit_message = "No commit message provided"

    commit_ts = _convert_time_to_commit_ts(commit_time)
    time_args = f'GIT_AUTHOR_DATE="{commit_ts.author_ts}" GIT_COMMITTER_DATE="{commit_ts.committer_ts}" '

    commit_message = shlex.quote(commit_message)
    no_changes_message = "No changes to commit"
    amend_args = "--amend " if amend else ""
    if allow_empty or amend:
        bash_command = f"""git add . && {time_args}git commit {amend_args}--allow-empty -m {commit_message} > /dev/null && git rev-parse HEAD"""
    else:
        bash_command = f"""git add . && ( git status | grep -q "nothing to commit" && echo "{no_changes_message}" ) || ( {time_args}git commit {amend_args}-m {commit_message} > /dev/null && git rev-parse HEAD )"""

    with Section(f"committing changes with message: '{commit_message}'", log_level="DEBUG"):
        stdout = await run_command_with_retry_on_git_lock_error(
            computing_environment,
            ["bash", "-c", bash_command],
        )
        stdout = stdout.strip()
        if stdout == no_changes_message:
            raise FailedToMakeCommitError(f"Failed to make commit with message: {commit_message}. {bash_command=}")
        new_git_hash = stdout
        return new_git_hash


async def get_tree_hash_for_commit(computing_environment: ComputingEnvironment, commit: str) -> str:
    """Get the tree hash for the given commit."""
    return await computing_environment.run_git(["rev-parse", commit + "^{tree}"])


async def get_commit_timestamp(computing_environment: ComputingEnvironment, commit: str) -> CommitTimestamp:
    """Get the commit timestamp for the given commit."""
    split_token = "<|>"
    result = await computing_environment.run_git(["show", "-s", "--format=%aI<|>%cI", commit])
    author_ts, committer_ts = result.split(split_token)
    return CommitTimestamp(author_ts=author_ts.strip(), committer_ts=committer_ts.strip())


async def tag_commit(computing_environment: ComputingEnvironment, tag: str, commit_hash: str) -> None:
    """Tag the given commit with the given tag."""
    # We use -f to force the tag to be created even if it already exists.
    await computing_environment.run_git(("tag", "-f", tag, commit_hash))


async def git_push(computing_environment: ComputingEnvironment, branch_name: str) -> str:
    """Push changes to remote branch with given name."""
    return await computing_environment.run_git(("push", "origin", branch_name))


async def force_push(computing_environment: ComputingEnvironment, branch_name: str) -> str:
    """Push changes to remote branch with given name."""
    return await computing_environment.run_git(("push", "--force", "origin", branch_name))


async def force_push_commit_with_retry(
    computing_environment: ComputingEnvironment, commit: str, branch_name: str, timeout: float = 30.0
) -> None:
    start_time = time.monotonic()
    sleep_time = 0.5
    while True:
        try:
            await force_push_commit(computing_environment, commit, branch_name)
            break
        except Exception as exc:
            if time.monotonic() - start_time > timeout:
                logger.error(f"Timeout reached: Could not force push {commit} to {branch_name} in {timeout} seconds.")
                raise TimeoutError(f"Timeout reached: Could not force push {commit} to {branch_name}.") from exc
            logger.info(f"Force push of {commit} to {branch_name} failed; trying again...")
            await asyncio.sleep(sleep_time)
            sleep_time *= 2


async def force_push_commit(computing_environment: ComputingEnvironment, commit: str, branch_name: str) -> None:
    try:
        await computing_environment.run_git(["push", "-f", "origin", f"{commit}:{branch_name}"], is_error_logged=False)
    except RunCommandError as e:
        if "fatal: bad object" in e.stderr:
            # TODO (danielmewes): We're retrying failed fetches here. However, there is also a separate
            #   force_push_commit_with_retry method that retries the entire force_push_commit.
            #   We should probably try the fetch only once, and then rely on the outer
            #   force_push_commit_with_retry to retry the entire force_push_commit call when retrying is =
            #   desired?
            NUM_TRIES = 3
            for _ in range(NUM_TRIES):
                try:
                    await computing_environment.run_git(["fetch", "origin", commit], is_error_logged=False)
                except RunCommandError as fetch_e:
                    if "not our ref" in fetch_e.stderr:
                        # FIXME: actually, this has been getting worse... I suspect perhaps rate limiting or something? We are checking thing out much more than usual...
                        await asyncio.sleep(2)
                    else:
                        raise fetch_e
                else:
                    start_time = time.monotonic()
                    while time.monotonic() - start_time < 10:
                        try:
                            await computing_environment.run_git(["push", "-f", "origin", f"{commit}:{branch_name}"])
                        except RunCommandError as repush_e:
                            if "not our ref" in repush_e.stderr:
                                # FIXME: actually, this has been getting worse... I suspect perhaps rate limiting or something? We are checking thing out much more than usual...
                                await asyncio.sleep(2)
                            else:
                                raise repush_e
                        else:
                            return
                    raise Exception(f"Could not force push commit {commit}")
            raise Exception(f"Could not fetch commit {commit} to force push it")
        else:
            raise


async def get_staged_files(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get list of all files in repo that are currently staged."""
    result = await computing_environment.run_git(("diff", "--full-index", "--binary", "--name-only", "--cached"))
    return tuple(result.splitlines())


async def get_unstaged_files(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get list of all files in repo that are currently unstaged."""
    result = await computing_environment.run_git(("diff", "--full-index", "--binary", "--name-only"))
    return tuple(result.splitlines())


async def restore_all_staged_files(computing_environment: ComputingEnvironment) -> None:
    """Restore all staged files."""
    await computing_environment.run_git(("restore", "--staged", "."))


async def restore_all_unstaged_changes(computing_environment: ComputingEnvironment) -> None:
    """Restore all unstaged changes."""
    await computing_environment.run_git(("restore", "."))


async def apply_patch_via_git_with_conflict_markers(
    computing_environment: ComputingEnvironment, git_diff: str, is_error_logged: bool = True
) -> None:
    """Apply a diff to repo with conflict markers."""
    if git_diff.strip() == "":
        return
    if not git_diff.endswith("\n"):
        # git requires a newline at the end of the patch
        git_diff += "\n"
    patch_file = _get_temp_patch_file()
    try:
        await computing_environment.write_file(patch_file, git_diff)
        await computing_environment.run_command(
            ["bash", "-c", f"git add . && git apply --verbose {patch_file} || git apply -3 --verbose {patch_file}"],
            is_error_logged=is_error_logged,
        )
    except RunCommandError as e:
        raise PatchApplicationError(f"Failed to apply patch: {e}") from e
    finally:
        await computing_environment.delete_file(patch_file)


async def is_repo_conflicted(computing_environment: ComputingEnvironment) -> bool:
    output = await computing_environment.run_git(["status"], is_error_logged=False, check=False)
    if "Unmerged paths:" in output:
        return True
    return False


async def get_head_hash(computing_environment: ComputingEnvironment) -> str:
    """Get the hash of the current HEAD commit."""
    git_hash = await computing_environment.run_git(["rev-parse", "HEAD"])
    assert len(git_hash) == 40, f"Expected 40-character git hash, got {git_hash}"
    return git_hash


async def get_parent_commit_hash(computing_environment: ComputingEnvironment, commit_hash: str) -> str:
    """Get the parent commit hash of the given commit hash."""
    git_hash = await computing_environment.run_git(["rev-parse", f"{commit_hash}^"])
    assert len(git_hash) == 40, f"Expected 40-character git hash, got {git_hash}"
    return git_hash


async def get_most_recent_sibling_branch_of_branch(
    computing_environment: ComputingEnvironment, target_branch_name: str
) -> Optional[Tuple[str, ...]]:
    """Get the most recent sibling branch of the given branch name.

    This is the first branch that shares a common ancestor commit with the given branch name.
    Note, that it is possible that there are multiple branches that share the most recent common ancestor.
    In this case, we return all of them.

    Also if there are no sibling branches (either at all or within the max lookback), we return None.
    """
    # FIXME: this is imprecise -- it really just needs to be head 2, but I didn't want to figure out the regex...
    output = await computing_environment.run_command(
        [
            "bash",
            "-c",
            f'git log --decorate=full --oneline {target_branch_name} | grep "refs/heads/" | head -n 10',
        ],
        check=True,
    )
    for line in output.splitlines():
        branch_list_string = line.split(" (", maxsplit=1)[-1].split(")", maxsplit=1)[0]
        parsed_branch_names = []
        for branch_name_string in branch_list_string.split(", "):
            parsed_branch_name = branch_name_string.rsplit(" ")[-1]
            if (
                parsed_branch_name.startswith("refs/heads")
                and parsed_branch_name != f"refs/heads/{target_branch_name}"
            ):
                parsed_branch_names.append(parsed_branch_name.replace("refs/heads/", "", 1))
        # otherwise, just return the first one
        if len(parsed_branch_names) > 0:
            return tuple(parsed_branch_names)
    return None


async def get_nth_commit_ago(
    computing_environment: ComputingEnvironment, commit_hash: str, n: int, is_error_logged: bool = True
) -> str:
    """Get the nth commit ago of the given commit hash."""
    git_hash = await computing_environment.run_git(
        ["rev-parse", f"{commit_hash}~{n}"], is_error_logged=is_error_logged
    )
    assert len(git_hash) == 40, f"Expected 40-character git hash, got {git_hash}"
    return git_hash


async def get_initial_repo_commit_hash(computing_environment: ComputingEnvironment, commit_hash: str = "HEAD") -> str:
    """Get the initial commit hash of the repo.

    As written, if invoked on an empty repo with no commits (immediately after `git init`), this fails with:
    `fatal: ambiguous argument 'HEAD': unknown revision or path not in the working tree.`
    """
    # --max-parents=0: only consider commits with no parents
    # --date-order: sort by date (newest first)
    output = await computing_environment.run_git(["rev-list", "--max-parents=0", commit_hash, "--date-order"])
    # assume the oldest commit with no parents is the initial repo commit
    all_root_commits = output.splitlines()
    root_commit = all_root_commits[-1]
    assert len(root_commit) == 40, f"Expected 40-character git hash, got {root_commit}"
    return root_commit


async def get_upto_nth_commit_ago(computing_environment: ComputingEnvironment, commit_hash: str, n: int) -> str:
    """Get the commit hash of the upto nth commit ago of the given commit hash.

    If the commit history is shorter than n, it will return the first commit.
    """
    try:
        return await get_nth_commit_ago(computing_environment, commit_hash, n, is_error_logged=False)
    except RunCommandError as e:
        if "unknown revision or path not in the working tree" in e.stderr:
            return await get_initial_repo_commit_hash(computing_environment, commit_hash)
        raise


async def get_commit_message(computing_environment: ComputingEnvironment, commit_hash: str) -> str:
    """Get the commit message of the given commit hash."""
    return await computing_environment.run_git(["log", "-1", "--pretty=%B", commit_hash])


async def get_commit_count_between_hashes(
    computing_environment: ComputingEnvironment, old_hash: str, new_hash: str
) -> int:
    """Get the number of commits between two hashes."""
    output = await computing_environment.run_git(["rev-list", "--count", f"{old_hash}..{new_hash}"])
    return int(output.strip())


async def fetch_and_checkout_hash(
    computing_environment: ComputingEnvironment, git_hash: str, is_error_logged: bool = True
) -> None:
    await computing_environment.run_command(
        ["bash", "-c", f"git fetch origin {git_hash} && git checkout {git_hash}"],
        is_error_logged=is_error_logged,
    )


async def fetch_ref_and_checkout_hash(
    computing_environment: ComputingEnvironment, ref: str, git_hash: str, is_error_logged: bool = True
) -> None:
    await computing_environment.run_command(
        ["bash", "-c", f"git fetch origin {ref} && git checkout {git_hash}"],
        is_error_logged=is_error_logged,
    )


@retry(
    stop=stop_after_attempt(5),
    wait=wait_random_exponential(min=0.5, max=30, exp_base=2),
    reraise=True,
    retry=retry_all(retry_if_exception_type(RunCommandError)),
    before_sleep=log_before_sleep,
)
async def wait_for_git_hash_to_checkout(computing_environment: ComputingEnvironment, git_hash: str) -> None:
    await fetch_and_checkout_hash(computing_environment, git_hash, is_error_logged=False)


async def wait_for_git_hash_with_ref_to_checkout(
    computing_environment: ComputingEnvironment, git_hash: str, ref: str, timeout: float = 20.0
) -> None:
    with Section(f"Checking out git hash {git_hash} with ref {ref}", log_level="DEBUG"):
        start_time = time.monotonic()
        while True:
            try:
                await fetch_ref_and_checkout_hash(computing_environment, ref, git_hash, is_error_logged=False)
                break
            except RunCommandError as exc:
                if time.monotonic() - start_time > timeout:
                    log_exception(
                        exc, f"Timeout reached: Git hash {git_hash} is not available after {timeout} seconds."
                    )
                    raise TimeoutError(f"Timeout reached: Git hash {git_hash} is not available.") from exc
                logger.info(f"Waiting for git hash {git_hash} to be available...")
                await asyncio.sleep(0.5)


async def force_checkout_git_hash_immediate_on_branch(
    computing_environment: ComputingEnvironment, git_hash: str, branch_name: str
) -> None:
    # Here we clear any uncommited changes, then change branch, before checking out the new hash
    # We need to do these three steps so we don't affect the currently checked out branch
    command = f"git reset --hard && git checkout -B {branch_name} && git reset --hard {git_hash}"
    logger.debug(f"Running command: {command}")
    await run_command_with_retry_on_git_lock_error(
        computing_environment,
        [
            "bash",
            "-c",
            command,
        ],
    )


async def get_git_folder_paths(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get the paths of all the git folders in the repo."""
    result = await computing_environment.run_command(["ls", ".git"])
    return tuple(result.splitlines())


async def apply_patch_via_git(
    computing_environment: ComputingEnvironment, git_diff: str, is_error_logged: bool
) -> None:
    """Apply a diff to repo."""
    if git_diff.strip() == "":
        return
    patch_file = _get_temp_patch_file()
    await computing_environment.write_file(patch_file, git_diff)
    # NOTE: --allow-empty is necessary because the patch may be empty, or result in no changes
    #  update (2024-11-22) --allow-empty is not available in the git version in our devcontainers
    #  so we have to do a janky error check below
    try:
        await computing_environment.run_git(("apply", "--verbose", str(patch_file)), is_error_logged=is_error_logged)
    except RunCommandError as e:
        raise PatchApplicationError(f"Failed to apply patch: {e}") from e
    finally:
        await computing_environment.delete_file(patch_file)


async def get_git_diff(
    computing_environment: ComputingEnvironment,
    commit_hash: str | None = None,
    staged: bool = False,
    is_error_logged: bool = True,
) -> str:
    """Get the diff for the current repo state."""
    # make sure `is_stripped=False` otherwise patch can be invalid
    command = ["diff", "--full-index", "--binary"]
    if staged:
        command.append("--staged")
    if commit_hash:
        command.append(commit_hash)
    return await computing_environment.run_git(command, is_stripped=False, is_error_logged=is_error_logged)


async def get_diff_between_hashes(computing_environment: ComputingEnvironment, old_hash: str, new_hash: str) -> str:
    """Get the diff between two git hashes."""
    # make sure `is_stripped=False` otherwise patch can be invalid
    return await computing_environment.run_git(
        ["diff", "--full-index", "--binary", old_hash, new_hash], is_stripped=False
    )


async def get_patch_for_commit(computing_environment: ComputingEnvironment, commit_hash: str) -> str:
    """Get the patch for a given commit hash."""
    return await computing_environment.run_git(["show", "--pretty=format:", "--patch", commit_hash], is_stripped=False)


async def get_untracked_files(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get the untracked files in the repo."""
    result = await computing_environment.run_git(["ls-files", "--others", "--exclude-standard"], is_error_logged=False)
    return tuple([line.strip() for line in result.splitlines() if line.strip()])


async def get_staged_unstaged_and_combined_diffs(
    computing_environment: ComputingEnvironment, base_commit: str = "HEAD"
) -> Tuple[str, str, str]:
    """Get the staged diff, the unstaged diff, and the combined diff"""
    staged_diff = await get_git_diff(computing_environment, staged=True)
    unstaged_diff = await get_git_diff(computing_environment, staged=False)
    combined_diff = await get_git_diff(computing_environment, commit_hash=base_commit)
    return staged_diff, unstaged_diff, combined_diff


async def get_unmerged_blob_hashes(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get the blob hashes of all the unmerged files in the repo."""
    result = await computing_environment.run_command(
        ["bash", "-c", "git ls-files --unmerged | awk '{print $2}' | sort -u"], check=False
    )
    return tuple(line.strip() for line in result.splitlines() if line.strip() != "")


async def get_staged_blob_hashes(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get the blob hashes of all the staged files in the repo."""
    staged_blob_hashes = await computing_environment.run_command(
        [
            "bash",
            "-c",
            'staged_blobs=$(git diff --full-index --binary --cached --name-only --diff-filter=ACMRT | while read file; do git ls-files --stage "$file" | awk \'{print $2}\'; done); echo "$staged_blobs"',
        ]
    )
    return tuple(line.strip() for line in staged_blob_hashes.splitlines() if line.strip() != "")


async def get_blob_content_by_hash(computing_environment: ComputingEnvironment, blob_hash: str) -> bytes:
    """Get the content of a blob by its hash."""
    result = await computing_environment.run_git(["cat-file", "-p", blob_hash], is_stripped=False)
    return result.encode("utf-8")


async def get_unmerged_and_staged_blob_contents_by_hash(
    computing_environment: ComputingEnvironment,
) -> Dict[str, bytes]:
    """Get the contents of all the unmerged and staged blobs in the repo."""
    unmerged_blob_hashes = await get_unmerged_blob_hashes(computing_environment)
    staged_blob_hashes = await get_staged_blob_hashes(computing_environment)
    all_relevant_blob_hashes = unmerged_blob_hashes + staged_blob_hashes
    blob_content_tasks_by_hash = {
        blob_hash: get_blob_content_by_hash(computing_environment, blob_hash) for blob_hash in all_relevant_blob_hashes
    }
    blob_contents = await asyncio.gather(*blob_content_tasks_by_hash.values())
    return {
        blob_hash: blob_content for blob_hash, blob_content in zip(blob_content_tasks_by_hash.keys(), blob_contents)
    }


async def write_blob_content(computing_environment: ComputingEnvironment, blob_hash: str, blob_content: bytes) -> None:
    """Write the content of a blob to the repo."""
    # write the blob content to a temp file
    temp_file = anyio.Path(f"/tmp/{blob_hash}")
    try:
        await computing_environment.write_file(temp_file, blob_content, mode="wb")
        # write the blob to the repo
        result = await computing_environment.run_git(["hash-object", "-w", str(temp_file)])
        assert result.strip() == blob_hash, f"Expected blob hash {blob_hash}, got {result.strip()}"
    finally:
        await computing_environment.delete_file(temp_file)


async def write_blob_content_by_hash(
    computing_environment: ComputingEnvironment, blob_content_by_hash: Dict[str, bytes]
) -> None:
    """Write the content of all the blobs to the repo."""
    tasks = []
    for blob_hash, blob_content in blob_content_by_hash.items():
        tasks.append(write_blob_content(computing_environment, blob_hash, blob_content))
    await asyncio.gather(*tasks)


async def get_modified_files_with_conflicts(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get the modified files with conflicts."""
    commands = ["diff --check --staged --full-index --binary", "diff --check --full-index --binary"]
    conflicted_files = set()
    for command in commands:
        result = await computing_environment.run_git(command.split(), check=False, is_error_logged=False)
        # output is of the form:
        # test.txt:2: leftover conflict marker

        for line in result.splitlines():
            parts = line.split(":", maxsplit=1)
            if len(parts) == 2:
                file_path, message = parts
                if "leftover conflict marker" in message.strip():
                    conflicted_files.add(file_path)
    return tuple(conflicted_files)


async def get_conflicted_pathnames(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get the pathnames of all the conflicted files in the repo."""
    result = await computing_environment.run_git(
        ["diff", "--full-index", "--binary", "--name-only", "--diff-filter=U"]
    )
    return tuple(result.splitlines())


async def get_conflicted_contents_by_path(computing_environment: ComputingEnvironment) -> Dict[str, bytes]:
    """Get the contents of all the conflicted files in the repo."""
    conflicted_files = await get_conflicted_pathnames(computing_environment)
    conflicted_contents_by_path: Dict[str, bytes] = {}
    for file_path in conflicted_files:
        content = await computing_environment.read_file(file_path, mode="rb")
        assert isinstance(content, bytes), f"Expected bytes, got {type(content)}"
        conflicted_contents_by_path[file_path] = content
    return conflicted_contents_by_path


async def get_modified_pathnames(computing_environment: ComputingEnvironment) -> Tuple[str, ...]:
    """Get the pathnames of all the modified files in the repo."""
    result = await computing_environment.run_command(["bash", "-c", "git status --porcelain | awk '{print $2}'"])
    return tuple(result.splitlines())


async def get_modified_file_contents_by_path(computing_environment: ComputingEnvironment) -> Dict[str, bytes]:
    """Get the contents of all the modified files in the repo."""
    modified_pathnames = await get_modified_pathnames(computing_environment)
    modified_file_contents_by_path: Dict[str, bytes] = {}
    for pathname in modified_pathnames:
        content = await computing_environment.read_file(pathname, mode="rb")
        assert isinstance(content, bytes), f"Expected bytes, got {type(content)}"
        modified_file_contents_by_path[pathname] = content
    return modified_file_contents_by_path


async def get_repo_url(computing_environment: ComputingEnvironment) -> str:
    repo_url = await computing_environment.run_git(["remote", "get-url", "origin"])
    if repo_url.startswith("git@"):
        # convert ssh url to https
        repo_url = repo_url.replace(":", "/")
        repo_url = f"https://{repo_url[4:]}"
    if "https://oauth2:" in repo_url:
        # remove the oauth2 prefix
        # repo_url is something like https://oauth2:{token}@gitlab.com/.../.git
        # change it to https://gitlab.com/.../.git
        suffix = repo_url.split("@")[-1]
        repo_url = "https://" + suffix
    return repo_url


async def get_main_branch_name_for_repo(
    computing_environment: ComputingEnvironment, default_branch: Optional[str] = None
) -> str:
    """Get the name of the main branch for the repo.

    Attempts to detect whether the repository uses 'main', 'master', or another name
    as its primary branch by checking for common branch names in order of preference.
    """
    possible_main_branches = ["main", "master", "trunk", "development"]

    if default_branch is not None and default_branch not in possible_main_branches:
        possible_main_branches.insert(0, default_branch)

    # First check if any of the common main branch names exist
    # and return the first one that does
    for branch in possible_main_branches:
        if await is_branch_exists(computing_environment, branch):
            return branch

    # If we couldn't find a common main branch, try to determine the default branch
    # This gets the branch that HEAD points to in a newly cloned repo
    default_remote_branch = await computing_environment.run_git(
        ["symbolic-ref", "refs/remotes/origin/HEAD"], is_error_logged=False
    )
    if default_remote_branch:
        # Format is typically refs/remotes/origin/main, so extract the last part
        default_remote_branch = default_remote_branch.strip().split("/")[-1]
        return default_remote_branch
    raise ValueError("Could not detect main branch for repo.")



from __future__ import annotations

import subprocess
from pathlib import Path
from typing import Any
from typing import Union

import anyio

# Use AnyPath type to match Sanctum
AnyPath = Union[Path, str, anyio.Path]


class RunCommandError(subprocess.CalledProcessError):
    """Custom exception for errors encountered during Git commands."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.cwd = kwargs.get("cwd", None)
        if "cwd" in kwargs:
            del kwargs["cwd"]
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        return f"Command `{self.cmd}` returned non-zero exit status {self.returncode}.\nOutput: {self.stdout}\nError: {self.stderr}\nCWD: {self.cwd}"


class PatchApplicationError(Exception):
    """Custom exception for errors encountered during patch application."""


class FailedToMakeCommitError(Exception):
    """Custom exception for errors encountered during commit creation."""



import os
import pathlib
from typing import Optional
from typing import Union


class YouAreBeingTooFancyInYourSettingsFile(Exception):
    pass


def parse_secrets_file(filepath: Union[str, pathlib.Path]) -> dict[str, str]:
    """Parse bashenv_secrets.sh-style file into a dict.
    We should REALLY NOT BE DOING THIS EVER but unfortunately that's not the case so at least let's only do it once here

    Not a great parser; will break in probably many scenarios but end-of-line comments are one that comes to mind
    """
    out: dict[str, str] = {}
    with open(filepath) as f:
        for line in f:
            if "$" in line:
                raise YouAreBeingTooFancyInYourSettingsFile(
                    "Yeah, don't do that. This .sh file is meant to be simple definitions, it should not use any features of bash or sh, including string interpolation via $"
                )
            if "#" in line:
                if not line.startswith("#"):
                    raise YouAreBeingTooFancyInYourSettingsFile("Put comments at the start of the line")
                continue
            if "\\" in line:
                raise YouAreBeingTooFancyInYourSettingsFile("No line continuations or other character escapes allowed")
            if line.startswith("export "):
                k, v = line.strip("export ").strip().split("=", maxsplit=1)
                k = k.strip()
                if k != k.upper():
                    raise YouAreBeingTooFancyInYourSettingsFile(f"Key {k} must be uppercase")
                v = v.strip()
                if v.startswith('"'):
                    if not v.endswith('"'):
                        raise YouAreBeingTooFancyInYourSettingsFile(f"Value {v} must end with a double quote")
                    v = v[1:-1]
                if v.startswith("'"):
                    if not v.endswith("'"):
                        raise YouAreBeingTooFancyInYourSettingsFile(f"Value {v} must end with a single quote")
                    v = v[1:-1]
                out[k] = v
            elif line.strip():
                raise YouAreBeingTooFancyInYourSettingsFile(
                    f"All lines must start with 'export ', but this line did not: {line}"
                )
    return out


# TODO: this is gross and bad--we should make better handling for secrets.
#  Right now we read the necessary secrets out of the bashenv files
def get_secret(secret_name: str) -> Optional[str]:
    value = os.environ.get(secret_name)
    if value is not None:
        return value
    secrets_files = (
        "science/secrets/environment_vars/bashenv.sh",
        "science/secrets/environment_vars/bashenv_secrets.sh",
        "science/secrets/environment_vars/common_vars.sh",
    )
    for file in secrets_files:
        if os.path.exists(file):
            secrets = parse_secrets_file(file)
            value = secrets.get(secret_name, None)
            if value is not None:
                return value
    return None



import contextlib
from typing import Iterable
from typing import Sequence
from typing import TypeVar

from imbue_core.errors import ImbueError

T = TypeVar("T")


class ImbueItertoolsValueError(ImbueError, ValueError):
    """This value error is thrown when the assumptions of the itertools module are violated."""


def only(x: Iterable[T]) -> T:
    try:
        (value,) = x
    except ValueError as e:
        message = "Expected exactly one value"
        if isinstance(x, Sequence):
            with contextlib.suppress():
                message += f" but got {len(x)} {x[:3]=}"
        raise ImbueItertoolsValueError(message) from e

    return value



import pytest
from pydantic import BaseModel

from imbue_core.pydantic_utils import model_update


def test_model_update_succeeds() -> None:
    class MyModel(BaseModel):
        field1: int
        field2: str

    original_model = MyModel(field1=1, field2="test")
    update = {"field1": 2}

    updated_model = model_update(original_model, update)

    assert updated_model.field1 == 2
    assert updated_model.field2 == "test"


def test_model_update_unknown_fields() -> None:
    class MyModel(BaseModel):
        field1: int
        field2: str

    original_model = MyModel(field1=1, field2="test")
    update = {"field4": 2}

    with pytest.raises(ValueError):
        updated_model = model_update(original_model, update)



import threading
from typing import Any
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import cast
from typing import get_args

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Discriminator
from pydantic import GetCoreSchemaHandler
from pydantic import Json
from pydantic.alias_generators import to_camel
from pydantic_core import core_schema as pyd_core_schema

from imbue_core.frozen_utils import FrozenDict
from imbue_core.nested_evolver import _Evolver
from imbue_core.nested_evolver import chill
from imbue_core.nested_evolver import evolver
from imbue_core.serialization_types import Serializable

T = TypeVar("T", bound=BaseModel)
V = TypeVar("V")

_threading_local = threading.local()


class EvolvableModel:
    def evolve(self: T, attribute: V, new_value: V) -> T:
        assert _threading_local.evolved_obj is not None, ".ref() must be called before evolve"

        assert isinstance(attribute, _Evolver)  # Tricked you, type system!
        dest_evolver: _Evolver[T] = cast(_Evolver[T], attribute)
        dest_evolver.assign(new_value)

        result = chill(_threading_local.evolved_obj)
        _threading_local.evolved_obj = None
        return result

    def ref(self: T) -> T:
        _threading_local.evolved_obj = evolver(self)
        return _threading_local.evolved_obj


class FrozenModel(EvolvableModel, BaseModel):
    """
    The base class for most internal data (that does not need to be serialized).

    We generally prefer to keep data immutable in order to avoid side effects, race conditions, etc
    """

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        arbitrary_types_allowed=False,
    )


class MutableModel(BaseModel):
    """
    The base class for any internal data that strictly must be mutable.  Should be used sparingly.
    """

    model_config = ConfigDict(
        frozen=False,
        extra="forbid",
        # FIXME: go back to preventing arbitrary types once we're done converting
        # arbitrary_types_allowed=False,
        arbitrary_types_allowed=True,
    )


class SerializableModel(EvolvableModel, BaseModel, Serializable):
    """
    The base class for all data that can be serialized to/from JSON.
    """

    model_config = ConfigDict(
        frozen=True,
        ser_json_bytes="base64",
        val_json_bytes="base64",
        alias_generator=to_camel,
        validate_by_alias=True,
        validate_by_name=True,
        # any extra values will end up in the __pydantic_extra__ field
        # this is effectively required for backwards compatibility
        # IMPORTANT: note that, by default, we clear this below!  These types are ONLY for backwards compatibility
        extra="allow",
        # this is also effectively required for backwards compatibility
        arbitrary_types_allowed=True,
    )

    # this is a place where we might way to do any backwards compatibility related logic
    def model_post_init(self, __context) -> None:
        self.__pydantic_extra__.clear()


def model_dump(obj: BaseModel, is_camel_case: bool = False) -> dict:
    return obj.model_dump(by_alias=is_camel_case)


def model_load(model_type: Type[T], data: dict) -> T:
    return model_type.model_validate(data)


def model_dump_json(obj: BaseModel | Json, is_camel_case: bool = False) -> str:
    return obj.model_dump_json(by_alias=is_camel_case)


def model_load_json(model_type: Type[T], data: str) -> T:
    return model_type.model_validate_json(data)


# this is mostly here for the default cases.
# When you want to upgrade a model (and keep it backwards compatible), you can make a custom discriminator
# (eg, that looks for the old type name or converts the old class names)
def build_discriminator(
    field_name: str = "object_type", additional_types_and_string_representations: Tuple[Tuple[Type, str], ...] = ()
) -> Discriminator:
    """
    Build a discriminator function for a Pydantic model.

    Args:
        field_name (str): The name of the field to use as the discriminator.
        additional_types_and_string_representations (Tuple[Tuple[Type, str], ...]): Register additional types to the discriminator.

    Returns:
        Callable[[T | dict], str]: A function that takes an instance of T or a dictionary and returns the value of the
            specified field.
    """

    def discriminator(obj: T | dict) -> str:
        for model_type, string_representation in additional_types_and_string_representations:
            if isinstance(obj, model_type):
                return string_representation
        if isinstance(obj, dict):
            return obj[field_name]
        return getattr(obj, field_name)

    return Discriminator(discriminator=discriminator)


class PydanticFrozenDictAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> pyd_core_schema.CoreSchema:
        def validate_from_dict(d: dict | FrozenDict) -> FrozenDict:
            return FrozenDict(d)

        frozendict_schema = pyd_core_schema.chain_schema(
            [
                handler.generate_schema(dict[*get_args(source_type)]),
                pyd_core_schema.no_info_plain_validator_function(validate_from_dict),
                pyd_core_schema.is_instance_schema(FrozenDict),
            ]
        )
        return pyd_core_schema.json_or_python_schema(
            json_schema=frozendict_schema,
            python_schema=frozendict_schema,
            serialization=pyd_core_schema.plain_serializer_function_ser_schema(dict),
        )



"""
Provides a context manager for logging a potentially time-consuming process, or a "section".

- Prints logs at start and end of a section.

- Prints a Markdown-like heading for nested sections: "#" for top-level sections, "##" for one level down, and so on.

- Emits structured logs for easier query.

The SectionWrapper context manager can be used to time the __enter__ and __exit__ methods of an existing context manager.
"""
import contextlib
import threading
import time
import typing
from asyncio import CancelledError
from types import TracebackType
from typing import Optional
from typing import Type
from typing import Union

from loguru import logger

_monotonic_base = time.monotonic()


def _monotonic_time() -> float:
    """A wrapper around time.monotonic() to make the return values a bit smaller and easier to read by a human."""
    return time.monotonic() - _monotonic_base


class _ThreadLocal(threading.local):
    def __init__(self) -> None:
        self.next_section_level: int = 0


_thread_local = _ThreadLocal()


class Section(contextlib.ContextDecorator):
    def __init__(self, message: str, log_level: Union[int, str] = "INFO") -> None:
        # TODO: loguru doesn't properly display integer log levels like e.g. logging.INFO
        self.message = message
        self.log_level = log_level

    def __enter__(self) -> "Section":
        level = _thread_local.next_section_level
        _thread_local.next_section_level += 1
        self.header = "#" * (level + 1)
        self.start_monotonic_time = _monotonic_time()
        start_clock_time = time.time()
        self.section = {
            "name": self.message,
            "level": level,
            "start_monotonic_time": self.start_monotonic_time,
            "start_clock_time": start_clock_time,
        }
        logger.log(
            self.log_level,
            f"{self.header} Start: {self.message}",
            section=self.section,
        )
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        _thread_local.next_section_level -= 1
        finish_monotonic_time = _monotonic_time()
        finish_clock_time = time.time()
        duration_seconds = finish_monotonic_time - self.start_monotonic_time
        section = self.section | {
            "finish_monotonic_time": finish_monotonic_time,
            "finish_clock_time": finish_clock_time,
            "duration_seconds": duration_seconds,
        }
        self.elapsed = duration_seconds

        if exc_val is None:
            logger.log(
                self.log_level,
                f"{self.header} Done: {self.message} (took {duration_seconds:.2f} seconds)",
                section=section | {"result": "success"},
            )
        else:
            if isinstance(exc_val, CancelledError):
                logger.log(
                    self.log_level,
                    f"{self.header} Cancelled: {self.message} (took {duration_seconds:.2f} seconds)",
                    section=section | {"result": "cancelled"},
                )
            else:
                logger.log(
                    self.log_level,
                    f"{self.header} Failed: {self.message} (within {duration_seconds:.2f} seconds)",
                    section=section | {"result": "failed"},
                )


T = typing.TypeVar("T")


class SectionWrapper(contextlib.AbstractContextManager[T]):
    def __init__(self, cm: contextlib.AbstractContextManager[T], enter_message: str, exit_message: str) -> None:
        self._cm = cm
        self._enter_message = enter_message
        self._exit_message = exit_message

    def __enter__(self) -> T:
        with Section(self._enter_message) as cm:
            return self._cm.__enter__()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        with Section(self._exit_message):
            self._cm.__exit__(exc_type, exc_val, exc_tb)



from typing import Annotated
from typing import Dict

from inline_snapshot import snapshot
from typing_extensions import TypeAliasType

from imbue_core.frozen_utils import FrozenDict
from imbue_core.frozen_utils import deep_freeze_mapping
from imbue_core.pydantic_serialization import PydanticFrozenDictAnnotation
from imbue_core.pydantic_serialization import SerializableModel
from imbue_core.pydantic_serialization import model_dump
from imbue_core.pydantic_serialization import model_dump_json
from imbue_core.pydantic_serialization import model_load
from imbue_core.pydantic_serialization import model_load_json


class TestObject(SerializableModel):
    name: str
    language_code: str
    inner_data: Dict[str, str]


def test_simple() -> None:
    obj = TestObject(**dict(name="Filiz", languageCode="tr-TR", innerData={"snake_key": "value", "camelKey": "value"}))
    assert model_dump(obj) == snapshot(
        {
            "name": "Filiz",
            "language_code": "tr-TR",
            "inner_data": {"snake_key": "value", "camelKey": "value"},
        }
    )


def test_to_camel() -> None:
    obj = TestObject(**dict(name="Filiz", languageCode="tr-TR", innerData={"snake_key": "value", "camelKey": "value"}))
    assert model_dump(obj, is_camel_case=True) == snapshot(
        {
            "name": "Filiz",
            "languageCode": "tr-TR",
            "innerData": {"snake_key": "value", "camelKey": "value"},
        }
    )


def test_reversible() -> None:
    obj = TestObject(**dict(name="Filiz", languageCode="tr-TR", innerData={"snake_key": "value", "camelKey": "value"}))
    assert model_load(TestObject, model_dump(obj)) == obj


class Example(SerializableModel):
    mapping: Annotated[FrozenDict[int | float, int | None | tuple[int, ...]], PydanticFrozenDictAnnotation]


def test_arbitrary_frozen_dict() -> None:
    x = deep_freeze_mapping(FrozenDict({0: 0, 0.10: None, 100: [1, 2, 3]}))
    obj = Example(mapping=x, mapping_orig=x)
    assert isinstance(obj.mapping, FrozenDict)
    assert obj.mapping == x
    assert obj.model_dump() == {"mapping": x}
    json = obj.model_dump_json()
    loaded = Example.model_validate_json(json)
    assert isinstance(loaded.mapping, FrozenDict)
    assert loaded.mapping == x


FrozenJSON = TypeAliasType(
    "FrozenJSON",
    "Union[FrozenDict[str, FrozenJSON], list[FrozenJSON], str, int, float, bool, None]",
)


class JsonExample(SerializableModel):
    mapping: Annotated[FrozenJSON, PydanticFrozenDictAnnotation]


def test_frozen_json() -> None:
    obj = JsonExample(
        mapping=FrozenDict(
            {
                "key1": FrozenDict({"key2": FrozenDict({"key3": 1})}),
                "key4": FrozenDict({"key5": FrozenDict({"key6": 2})}),
                "list_key": [1, 2, 4.4, "string", True],
            }
        )
    )
    assert model_load_json(JsonExample, model_dump_json(obj)) == obj


def test_evolve() -> None:
    obj = TestObject(
        name="Filiz",
        language_code="tr-TR",
        inner_data={"snake_key": "value", "camelKey": "value"},
    )
    new_obj = obj.evolve(obj.ref().name, "thing")
    assert new_obj.name == "thing"
    assert obj.name == "Filiz"



from datetime import datetime
from typing import Annotated

from pydantic.functional_validators import PlainValidator

from imbue_core.pydantic_serialization import FrozenModel
from imbue_core.pydantic_serialization import SerializableModel


def _validate_git_timestamp(value: str) -> str:
    try:
        datetime.fromisoformat(value)
        return value
    except ValueError:
        raise ValueError(f"Invalid git timestamp: {value}")


class CommitTimestamp(SerializableModel):
    author_ts: Annotated[str, PlainValidator(_validate_git_timestamp)]
    committer_ts: Annotated[str, PlainValidator(_validate_git_timestamp)]


class CommitMetadata(FrozenModel):
    commit: str
    tree_hash: str
    message: str
    commit_time: CommitTimestamp

    @property
    def body(self) -> str:
        return self.message.split("\n", 1)[-1]

    @property
    def subject(self) -> str:
        return self.message.split("\n", 1)[0]



from __future__ import annotations

import asyncio
import typing
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Union

import loguru

if typing.TYPE_CHECKING:
    loguru_record = loguru.Record
else:
    loguru_record = Dict[str, Any]
FilterDict = Dict[Union[str, None], Union[str, int, bool]]
FilterFunction = Callable[[loguru_record], bool]
LOCATION_WIDTH = 60
TRACE = "TRACE"

TASK_ID_MESSAGE_WIDTH = 12


def fix_full_location(record: "loguru.Record") -> str:
    """
    One goal of this function is to format the location in an IDE-friendly way,
    so that control-clicking on the logged location opens the correct file
    and puts the cursor at the correct line.

    `record` looks like this:
    ```
    {
        "elapsed": datetime.timedelta(seconds=5, microseconds=152312),
        "exception": None,
        "extra": {
            "machine": "32de5bcafaa8",
            "user": "user",
            "agent_type": None,
            "agent_id": None,
            "parent_id": None,
            "async_task_id": None,
            "formatted_task_id": "",
            "formatted_agent_id": "",
            "sandbox_id": None,
            "formatted_sandbox_id": "",
        },
        "file": (
            name="error_dump_utils.py",
            path="/thad/dropbox/Thad Hughes/src/generally_intelligent/computronium/computronium/common/error_dump_utils.py",
        ),
        "function": "write_exception",
        "level": (name="INFO", no=20, icon="ℹ️"),
        "line": 214,
        "message": "Full traceback for  is available at http://node-004.snake-blues.ts.net:7777/exceptions/thad__notebook_2025_01_28_legolas/1000__2025_01_28_17_47_19_170102__P000_W0000/C0000__user_100.110.58.95_5045/2025_03_06_09_59_10_188008_KeyboardInterrupt_traceback.txt",
        "module": "error_dump_utils",
        "name": "computronium.common.error_dump_utils",
        "process": (id=3099015, name="MainProcess"),
        "thread": (id=140434013706048, name="MainThread"),
        "time": datetime(
            2025, 3, 6, 9, 59, 10, 920101, tzinfo=datetime.timezone(datetime.timedelta(days=-1, seconds=57600), "PST")
        ),
    }
    ```
    """
    log_path = Path(record["file"].path)
    if log_path.is_relative_to(Path.cwd()):
        log_path = log_path.relative_to(Path.cwd())
    location = record["extra"].get("full_location", f'{str(log_path)}:{record["line"]}:{record["function"]}')
    while len(location) > LOCATION_WIDTH and "/" in location:
        location = location[location.find("/") + 1 :]
    return location[-LOCATION_WIDTH:].rjust(LOCATION_WIDTH)


def format_task_id(async_task_id: str) -> str:
    return async_task_id[:TASK_ID_MESSAGE_WIDTH].rjust(TASK_ID_MESSAGE_WIDTH)


def patch_log_context_in_place(record: "loguru.Record", format_task_id: Callable[[str], str] = format_task_id) -> None:
    record["extra"]["full_location"] = fix_full_location(record)

    async_task_id = None
    try:
        # get the task id
        current_task = asyncio.current_task()
        if current_task is not None:
            async_task_id = current_task.get_name()
    except RuntimeError:
        # we're not in an asyncio event loop
        pass

    if async_task_id:
        formatted_task_id = format_task_id(async_task_id)
        record["extra"]["formatted_task_id"] = f" [{formatted_task_id}]"
        record["extra"]["async_task_id"] = async_task_id



from imbue_core.frozen_utils import FrozenDict
from imbue_core.frozen_utils import JSON
from imbue_core.frozen_utils import deep_freeze_json


def test_frozen_dict_hash() -> None:
    """Ensure that FrozenDict hashes and compares correctly."""
    assert FrozenDict() == FrozenDict()
    assert hash(FrozenDict()) == hash(FrozenDict())

    a = FrozenDict({1: 1, 2: 2})
    b = FrozenDict({2: 2, 1: 1})
    assert len({a, b}) == 1
    assert tuple(a) != tuple(b)


def test_deep_freeze_json() -> None:
    """Ensure that deep_freeze_json exhibits reasonable behavior."""
    json: JSON = {
        "a": [1, 2, 3],
        "b": {
            "c": 4,
            "d": {
                "e": [5, 6, 7],
            },
        },
        "c": False,
        "d": None,
    }

    assert deep_freeze_json(json) == FrozenDict(
        {
            "a": (1, 2, 3),
            "b": FrozenDict(
                {
                    "c": 4,
                    "d": FrozenDict(
                        {
                            "e": (5, 6, 7),
                        }
                    ),
                }
            ),
            "c": False,
            "d": None,
        }
    )



import asyncio
import sys
import traceback
from asyncio import Future
from types import TracebackType
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Type

import sentry_sdk

from imbue_core.constants import ExceptionPriority
from imbue_core.error_utils import timeout_get_traceback_with_vars
from imbue_core.errors import ExpectedError

_IS_SHUTTING_DOWN = False


# This is the name of the attribute we set on our exceptions to ensure they are logged (esp. to Sentry) at most once.
EXCEPTION_LOGGED_FLAG = "_was_logged_by_log_exception"


def notify_task_groups_of_shutdown() -> None:
    global _IS_SHUTTING_DOWN
    _IS_SHUTTING_DOWN = True


class PropagatingTaskGroup(asyncio.TaskGroup):
    """Improves over TaskGroup by ensuring that cancelation messages are actually propagated"""

    def __init__(self) -> None:
        # deferring the import in case this doesn't get used
        pass

        python_version: sys._version_info = sys.version_info
        if python_version[:2] != (3, 11) and python_version[:2] != (3, 12):
            raise RuntimeError(
                f"Python version 3.11 or 3.12 is required. You are using {python_version.major}.{python_version.minor}"
            )
        super().__init__()
        self._entered = False
        self._exiting = False
        self._aborting = False
        self._loop = None
        self._parent_task: Optional[asyncio.Task] = None
        self._parent_cancel_requested = False
        self._tasks: Set[asyncio.Task] = set()
        self._errors: List[BaseException] = []
        self._base_error: Optional[BaseException] = None
        self._on_completed_fut: Optional[Future[Any]] = None
        self._original_message: Optional[str] = None

    async def __aexit__(
        self, et: Optional[Type[BaseException]], exc: Optional[BaseException], tb: Optional[TracebackType]
    ) -> None:
        tb = None
        try:
            return await self._aexit(et, exc)
        finally:
            # Exceptions are heavy objects that can have object
            # cycles (bad for GC); let's not keep a reference to
            # a bunch of them. It would be nicer to use a try/finally
            # in __aexit__ directly but that introduced some diff noise
            self._parent_task = None
            self._errors = None
            self._base_error = None
            exc = None

    async def _aexit(self, et: Optional[Type[BaseException]], exc: Optional[BaseException]) -> None:
        self._exiting = True

        if exc is not None and self._is_base_error(exc) and self._base_error is None:  # type: ignore
            self._base_error = exc

        propagate_cancellation_error = exc if et is asyncio.exceptions.CancelledError else None
        if self._parent_cancel_requested:
            # If this flag is set we *must* call uncancel().
            assert self._parent_task
            if self._parent_task.uncancel() == 0:
                # If there are no pending cancellations left,
                # don't propagate CancelledError.
                propagate_cancellation_error = None

        if et is not None:
            if not self._aborting:
                # Our parent task is being cancelled:
                #
                #    async with TaskGroup() as g:
                #        g.create_task(...)
                #        await ...  # <- CancelledError
                #
                # or there's an exception in "async with":
                #
                #    async with TaskGroup() as g:
                #        g.create_task(...)
                #        1 / 0
                assert exc is not None
                self._abort_and_propagate(exc)

        # We use while-loop here because "self._on_completed_fut"
        # can be cancelled multiple times if our parent task
        # is being cancelled repeatedly (or even once, when
        # our own cancellation is already in progress)
        while self._tasks:
            if self._on_completed_fut is None:
                assert self._loop
                self._on_completed_fut = self._loop.create_future()

            try:
                await self._on_completed_fut
            except asyncio.exceptions.CancelledError as ex:
                if not self._aborting:
                    # Our parent task is being cancelled:
                    #
                    #    async def wrapper():
                    #        async with TaskGroup() as g:
                    #            g.create_task(foo)
                    #
                    # "wrapper" is being cancelled while "foo" is
                    # still running.
                    propagate_cancellation_error = ex
                    self._abort_and_propagate(ex)

            self._on_completed_fut = None

        assert not self._tasks

        if self._base_error is not None:
            try:
                raise self._base_error
            finally:
                exc = None

        # Propagate CancelledError if there is one, except if there
        # are other errors -- those have priority.
        try:
            if propagate_cancellation_error and not self._errors:
                try:
                    raise propagate_cancellation_error
                finally:
                    exc = None
        finally:
            propagate_cancellation_error = None

        if et is not None and et is not asyncio.exceptions.CancelledError:
            assert exc is not None
            self._errors.append(exc)

        if self._errors:
            try:
                raise BaseExceptionGroup(
                    "unhandled errors in a TaskGroup: see earlier in logs for causal error!",
                    self._errors,
                ) from None
            finally:
                exc = None

    def _abort(self) -> None:
        raise Exception("Please call _abort_and_propagate instead")

    def _abort_and_propagate(self, exc: BaseException) -> None:
        global _IS_SHUTTING_DOWN
        self._aborting = True

        if self._original_message is None:
            if isinstance(exc, asyncio.exceptions.CancelledError) and len(exc.args) > 0:
                self._original_message = "TaskGroup canceled because:\n" + exc.args[0]
            else:
                if not isinstance(exc, asyncio.exceptions.CancelledError):
                    if not _IS_SHUTTING_DOWN and not isinstance(exc, ExpectedError):
                        log_exception(exc, "Emergency print of error that caused task group to die:")
                self._original_message = f"TaskGroup died because: {type(exc).__name__}: {exc}\n" + "".join(
                    traceback.extract_tb(exc.__traceback__).format()
                )

        for t in self._tasks:
            if not t.done():
                t.cancel(self._original_message)

    def _on_task_done(self, task: asyncio.Task) -> None:
        self._tasks.discard(task)

        if self._on_completed_fut is not None and not self._tasks:
            if not self._on_completed_fut.done():
                self._on_completed_fut.set_result(True)

        if task.cancelled():
            return

        exc = task.exception()
        if exc is None:
            return

        self._errors.append(exc)
        if self._is_base_error(exc) and self._base_error is None:  # type: ignore
            self._base_error = exc

        assert self._parent_task
        if self._parent_task.done():
            # Not sure if this case is possible, but we want to handle
            # it anyways.
            assert self._loop
            self._loop.call_exception_handler(
                {
                    "message": f"Task {task!r} has errored out but its parent "
                    f"task {self._parent_task} is already completed",
                    "exception": exc,
                    "task": task,
                }
            )
            return

        if not self._aborting and not self._parent_cancel_requested:
            # If parent task *is not* being cancelled, it means that we want
            # to manually cancel it to abort whatever is being run right now
            # in the TaskGroup.  But we want to mark parent task as
            # "not cancelled" later in __aexit__.  Example situation that
            # we need to handle:
            #
            #    async def foo():
            #        try:
            #            async with TaskGroup() as g:
            #                g.create_task(crash_soon())
            #                await something  # <- this needs to be canceled
            #                                 #    by the TaskGroup, e.g.
            #                                 #    foo() needs to be cancelled
            #        except Exception:
            #            # Ignore any exceptions raised in the TaskGroup
            #            pass
            #        await something_else     # this line has to be called
            #                                 # after TaskGroup is finished.
            self._abort_and_propagate(exc)
            self._parent_cancel_requested = True
            self._parent_task.cancel(self._original_message)


def safe_cancel(task: asyncio.Task, msg: Optional[str] = None) -> None:
    """
    NOTE: this is probably not what you want!  See safe_cancel_and_wait_for_cleanup below for the more common use case.

    Cancels a task in a way that preserves information about who canceled it.

    Without using this, it is super obnoxious to figure out why your function is being canceled --
    you just get a CancelledError with no traceback.

    We try to ensure that *all* of our tasks are canceled in this way, which makes debugging much easier.

    Even safe_cancel_and_wait_for_cleanup will cancel in this way. The only difference is that that function
    also waits for the task to actually be canceled. Otherwise, cancellation just enqueues a cancellation.

    Note also that cancellation is never guaranteed -- all it does is raise a CancelledError in the task.
    This is why it is so important to never swallow those errors!
    """
    task.is_being_canceled_by_us = True  # type: ignore
    message = f"Task canceled by: \n {''.join(traceback.format_stack()[:-1])}"
    if msg:
        message += f"\nOriginal message: {msg}"

    task.cancel(message)


async def safe_cancel_and_wait_for_cleanup(
    task: asyncio.Task,
    msg: Optional[str] = None,
    exception_types_to_ignore: Sequence[Type[BaseException]] = (),
) -> None:
    """
    Convenience function for calling safe_cancel_multiple_and_wait_for_cleanup with a single task.

    See safe_cancel_multiple_and_wait_for_cleanup for docs.
    """
    await safe_cancel_multiple_and_wait_for_cleanup([task], msg, exception_types_to_ignore)


async def safe_cancel_multiple_and_wait_for_cleanup(
    tasks: Sequence[asyncio.Task],
    msg: Optional[str] = None,
    exception_types_to_ignore: Sequence[Type[BaseException]] = (),
) -> None:
    """
    Calls safe_cancel (see docs above) on each task in tasks, then waits for them to be done.

    Note that you can pass in a list of exception types to ignore.
    This is important for suppressing exceptions from third party libraries.
    You should probably make a constant in your project that lists these exceptions.

    We cannot simply suppress all BaseExceptions here because you really don't want to do that for things like signals and OutOfMemoryError
    """
    for task in tasks:
        safe_cancel(task, msg)
    # if you really want something to be canceled, you need to wait for it to be done
    # https://docs.python.org/3/library/asyncio-task.html#asyncio.Task.cancel
    results = await asyncio.gather(*tasks, return_exceptions=True)
    exceptions = []
    for result in results:
        if isinstance(result, BaseException):
            if isinstance(result, asyncio.CancelledError):
                pass
            elif isinstance(result, ExceptionGroup):
                exceptions.extend(_filter_exception_group(result))
            else:
                exceptions.append(result)  # type: ignore
        # cannot do this because the task may have just finished
        # assert (
        #     False
        # ), f"While cancelling async task and waiting for cleanup, expected None or CancelledError, got `{type(x)}: {x}`"

    filtered_exceptions = []
    for exception in exceptions:
        if not any(isinstance(exception, exception_type) for exception_type in exception_types_to_ignore):
            filtered_exceptions.append(exception)

    if len(filtered_exceptions) == 1:
        raise filtered_exceptions[0]
    elif len(filtered_exceptions) > 1:
        raise ExceptionGroup("Multiple exceptions in task group while canceling", filtered_exceptions)


def _filter_exception_group(exc_group: ExceptionGroup) -> List[Exception]:
    """Recursively extract exceptions from ExceptionGroups (ignoring canceled errors)."""
    result = []
    for exc in exc_group.exceptions:
        if isinstance(exc, asyncio.CancelledError):
            continue
        elif isinstance(exc, ExceptionGroup):
            result.extend(_filter_exception_group(exc))
        else:
            result.append(exc)
    return result


def log_exception(
    exc: BaseException,
    message: str,
    priority: Optional[ExceptionPriority] = None,
    *args,
    **kwargs,
) -> None:
    """Josh doesn't like that `loguru.exception()` takes only a message, and grabs the current exception from sys.exc_info().

    So this is a more explicit alternative that takes the exception as an argument.
    """
    # deferred import, will have been imported anyway by this point
    from loguru import logger

    if getattr(exc, EXCEPTION_LOGGED_FLAG, False):
        logger.info("Skipping duplicate log of exception {} with message {!r}", exc, message)
        return

    try:
        setattr(exc, EXCEPTION_LOGGED_FLAG, True)
    except AttributeError:
        logger.info("Unable to guarantee that {} will not be logged again", exc)

    # use a new scope to ensure these attachments don't bleed to other events that might have the same scope
    with sentry_sdk.new_scope() as scope:
        traceback_str = "".join(traceback.format_stack())
        message = f"{message}\n\nlog_exception CALL SITE TRACEBACK:\n\n{traceback_str}\nLOGGED EXCEPTION TRACEBACK FOLLOWS:\n"

        # attach traceback of log_exception callsite
        scope.add_attachment(bytes=traceback_str.encode(), filename="log_exception_traceback.txt")

        # for original exception, get traceback with variables and attach
        traceback_with_variables = timeout_get_traceback_with_vars(exc)
        if traceback_with_variables is not None:
            scope.add_attachment(
                bytes=traceback_with_variables.encode(), filename="original_exception_traceback_with_vars.txt"
            )

        # inject received exception stack trace into logger error message
        options = (exc,) + logger._options[1:]
        if priority is not None:
            level = priority.value
        else:
            level = "ERROR"
        logger._log(level, False, options, message, args, kwargs)


def apply() -> None:
    asyncio.TaskGroup = PropagatingTaskGroup  # type: ignore
    asyncio.taskgroups.TaskGroup = PropagatingTaskGroup  # type: ignore



"""Utility abstractions for interacting with git repositories."""
from __future__ import annotations

import asyncio
import contextlib
import shlex
import shutil
import subprocess
import sys
from asyncio.subprocess import PIPE
from asyncio.subprocess import STDOUT
from contextlib import asynccontextmanager
from io import StringIO
from pathlib import Path
from types import TracebackType
from typing import Any
from typing import AsyncGenerator
from typing import AsyncIterator
from typing import Dict
from typing import List
from typing import Optional
from typing import Self
from typing import Sequence
from typing import TYPE_CHECKING
from typing import TextIO
from typing import Tuple
from typing import Type
from typing import Union

import anyio
import attr
from loguru import logger

from imbue_core.async_monkey_patches import log_exception
from imbue_core.async_utils import sync
from imbue_core.async_utils import sync_contextmanager_func
from imbue_core.computing_environment.computing_environment import assert_repo_is_clean
from imbue_core.computing_environment.computing_environment import get_head_hash
from imbue_core.computing_environment.computing_environment import git_add
from imbue_core.computing_environment.computing_environment import is_repo_dirty
from imbue_core.computing_environment.computing_environment import make_commit
from imbue_core.computing_environment.computing_environment import restore_all_staged_files
from imbue_core.computing_environment.computing_environment import restore_all_unstaged_changes
from imbue_core.computing_environment.computing_environment import run_command_with_retry_on_git_lock_error
from imbue_core.computing_environment.data_types import AnyPath
from imbue_core.computing_environment.data_types import RunCommandError

if TYPE_CHECKING:
    # for proper file mode typing
    from _typeshed import OpenBinaryMode
    from _typeshed import OpenBinaryModeReading
    from _typeshed import OpenBinaryModeWriting
    from _typeshed import OpenTextMode
    from _typeshed import OpenTextModeReading
    from _typeshed import OpenTextModeWriting

PYTHON_EXTENSION = ".py"


def is_path_in_git_repo(path: Path) -> bool:
    """Check if a path is in a git repository."""
    if path.is_file():
        path = path.parent
    completed_process = subprocess.run(
        ["git", "-C", path, "rev-parse", "--is-inside-work-tree"],
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
    )
    if completed_process.returncode != 0:
        return False
    result = completed_process.stdout.decode().strip()
    assert result in ("true", "false"), result
    return result == "true"


def get_git_repo_root() -> Path:
    """Gets a Path to the current git repo root, assuming that our cwd is somewhere inside the repo."""
    completed_process = subprocess.run(
        ("git", "rev-parse", "--show-toplevel"),
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )
    root_dir = Path(completed_process.stdout.decode().strip())
    assert root_dir.is_dir(), f"{root_dir} must be a directory"
    return root_dir


def get_git_repo_root_from_path(path: Path) -> Path:
    """Gets a Path to the git repo root for the given path."""
    if path.is_file():
        path = path.parent
    completed_process = subprocess.run(
        ["git", "-C", path, "rev-parse", "--show-toplevel"],
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )
    root_dir = Path(completed_process.stdout.decode().strip())
    assert root_dir.is_dir(), f"{root_dir} must be a directory"
    return root_dir


REPO_LOCKS: Dict[Path, asyncio.Lock] = {}


@attr.s(auto_attribs=True, frozen=True)
class LocalGitRepo:
    base_path: Path

    @classmethod
    def build_from_cwd(cls) -> Self:
        """Create a `LocalGitRepo` instance from the current working directory."""
        return cls(get_git_repo_root())

    async def run_git(
        self,
        command: Sequence[str],
        check: bool = True,
        cwd: Optional[AnyPath] = None,
        is_error_logged: bool = True,
        is_stripped: bool = True,
        retry_on_git_lock_error: bool = True,
    ) -> str:
        """Run a git command in the repo.

        Example:
        ```
        git_repo.run_git("status")
        ```
        """
        # TODO: check for whether hooks should actually be run when we call this function
        absolute_path = self.base_path.absolute()
        if absolute_path not in REPO_LOCKS:
            REPO_LOCKS[absolute_path] = asyncio.Lock()
        async with REPO_LOCKS[absolute_path]:
            command = ["git"] + list(command)
            if not retry_on_git_lock_error:
                result = await self.run_command(command, check=check, is_error_logged=is_error_logged, cwd=cwd)
            else:
                result = await run_command_with_retry_on_git_lock_error(
                    self, command, check=check, is_error_logged=is_error_logged, cwd=cwd
                )
            if is_stripped:
                return result.strip()
            return result

    sync_run_git = sync(run_git)

    async def run_command(
        self,
        command: Sequence[str],
        check: bool = True,
        secrets: Optional[Dict[str, str]] = None,
        cwd: Optional[AnyPath] = None,
        is_error_logged: bool = True,
    ) -> str:
        """Run a command in the repo.

        Note, this can be used to run any command, not just git.
        """
        command_string = shlex.join(command)
        logger.trace(
            f"Running command: {command_string=} from cwd={cwd or self.base_path} with {secrets=} {check=} {is_error_logged=}"
        )
        proc = await asyncio.create_subprocess_exec(
            *command,
            cwd=cwd or self.base_path,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=secrets,
        )
        # note, need to be carefull not to strip() lines since whitespace may be important (e.g. for diffs)
        # return joined lines since mostly we only use the output for logging, and this way we arn't
        # passing around lots of lists. Also it's easy to parse by lines if needed
        stdout_bytes, stderr_bytes = await proc.communicate()
        try:
            stdout = stdout_bytes.decode("UTF-8")
        except UnicodeDecodeError as e:
            # If we don't encounter this, it likely means something was fixed upstream and we can safely delete
            log_exception(
                e,
                f"Command {command_string} failed to decode stdout, replacing any invalid bytes which could lead to problems later",
            )
            stdout = stdout_bytes.decode("UTF-8", errors="replace")
        stderr = stderr_bytes.decode("UTF-8")
        if check and proc.returncode != 0:
            error_message = f"command run from cwd={self.base_path} failed with exit code {proc.returncode} and stdout:\n{stdout}\nstderr:\n{stderr}"
            if is_error_logged:
                logger.error(
                    f"command attempted: '{command_string}' from cwd={self.base_path}\nerror message: {error_message}"
                )
            # this should not be None, but do this to satisfy type checker, int or None we throw the same error
            returncode = proc.returncode or -1
            raise RunCommandError(
                cmd=command_string,
                stderr=stderr,
                returncode=returncode,
                cwd=cwd or self.base_path,
            )
        return stdout

    @contextlib.asynccontextmanager
    async def _open_file(
        self,
        relative_path: AnyPath,
        cwd: Optional[AnyPath] = None,
        mode: Union[OpenTextMode, OpenBinaryMode] = "r",
        mkdir_if_missing: bool = True,
    ) -> AsyncGenerator[anyio.AsyncFile[Any], None]:
        logger.trace(f"opening file {relative_path} in cwd {cwd} with mode {mode}")
        if cwd is not None:
            sb_file_path = str(Path(cwd) / relative_path)
        else:
            sb_file_path = str(self.base_path / relative_path)

        if mkdir_if_missing:
            parent_dir = anyio.Path(sb_file_path).parent
            await parent_dir.mkdir(parents=True, exist_ok=True)

        f: Optional[anyio.AsyncFile[Any]] = None
        try:
            f = await anyio.Path(sb_file_path).open(mode=mode)  # type: ignore
            yield f
        finally:
            if f is not None:
                await f.aclose()

    async def write_file(
        self,
        relative_path: AnyPath,
        content: Optional[Union[str, bytes]],
        cwd: Optional[AnyPath] = None,
        mode: Union[OpenTextModeWriting, OpenBinaryModeWriting] = "w",
        mkdir_if_missing: bool = True,
    ) -> None:
        if content is None:
            await self.delete_file(relative_path, cwd=cwd)
            return

        async with self._open_file(relative_path, cwd=cwd, mode=mode, mkdir_if_missing=mkdir_if_missing) as f:
            logger.trace(f"writing to file {relative_path} in cwd {cwd} with mode {mode}")
            await f.write(content)

    async def delete_file(self, relative_path: AnyPath, cwd: Optional[AnyPath] = None) -> None:
        logger.trace(f"deleting the file {relative_path} in cwd {cwd}")
        if cwd is not None:
            sb_file_path = str(Path(cwd) / relative_path)
        else:
            sb_file_path = str(self.base_path / relative_path)
        await anyio.Path(sb_file_path).unlink()

    async def read_file(
        self,
        relative_path: AnyPath,
        cwd: Optional[AnyPath] = None,
        mode: Union[OpenTextModeReading, OpenBinaryModeReading] = "r",
        mkdir_if_missing: bool = True,
    ) -> Union[str, bytes]:
        async with self._open_file(relative_path, cwd=cwd, mode=mode, mkdir_if_missing=mkdir_if_missing) as f:
            logger.trace(f"reading file {relative_path} in cwd {cwd} with mode {mode}")
            content = await f.read()
            assert isinstance(content, str) or isinstance(content, bytes)
            return content

    async def head_hash(self) -> str:
        """Get the hash of the current HEAD commit."""
        return await get_head_hash(self)

    async def is_git_repo(self) -> bool:
        """Check that repo is valid git repo."""
        return await anyio.Path(self.base_path / ".git").exists()

    sync_is_git_repo = sync(is_git_repo)

    async def assert_clean(self) -> None:
        await assert_repo_is_clean(self)

    sync_assert_clean = sync(assert_clean)

    async def configure_git(
        self,
        git_user_name: Optional[str] = None,
        git_user_email: Optional[str] = None,
        initial_commit_message: str = "initial commit",
        is_recreating: bool = False,
    ) -> None:
        """Configure git repo with user name and email."""
        if is_recreating:
            if await self.is_git_repo():
                await asyncio.to_thread(shutil.rmtree, self.base_path / ".git")

        # order here is important
        # ref https://stackoverflow.com/questions/11656761/git-please-tell-me-who-you-are-error?noredirect=1
        await self.run_git(("init",))
        if git_user_name:
            await self.run_git(("config", "user.name", f"'{git_user_name}'"))
        if git_user_email:
            await self.run_git(("config", "user.email", f"'{git_user_email}'"))
        await self.run_git(("add", "."))
        await self.run_git(("commit", "-m", f"'{initial_commit_message}'"))
        branch_name = await self.run_git(("symbolic-ref", "HEAD"))
        if not branch_name == "refs/heads/main":
            # rename master to main for consistency
            await self.run_git(("branch", "-m", "master", "main"))

    sync_configure_git = sync(configure_git)

    @asynccontextmanager
    async def temporary_commit(
        self, tag_prefix: str, commit_message: str, raise_on_head_hash_change: bool = False
    ) -> AsyncIterator[str]:
        """Context manager to make a temporary commit and tag in the repo."""
        await self.run_git(("commit", "-am", commit_message, "--allow-empty", "--no-verify"))
        head_hash = await self.head_hash()
        tag = f"{tag_prefix}/{head_hash}"
        await self.run_git(("tag", tag))
        await self.run_git(("push", "origin", tag, "--no-verify"))
        try:
            yield head_hash
        finally:
            # This is susceptible to a race condition (if the user makes a commit between the time we check the head hash and the time we reset the state).
            # So it's important to keep any block that uses this context manager short - make the commit, copy it to the controller, and work there. Don't hold the repo hostage.
            current_head_hash = await self.head_hash()
            if current_head_hash != head_hash and raise_on_head_hash_change:
                raise AssertionError(
                    f"Head hash has changed from {head_hash} to {current_head_hash} since the temporary commit was made. Giving up on resetting git state, please address this manually."
                )
            else:
                await self.run_git(("reset", "HEAD~"))

    sync_temporary_commit = sync_contextmanager_func(temporary_commit)

    async def copy_repo(self, new_repo_path: Path, exists_ok: bool = True) -> "LocalGitRepo":
        """Make a full copy of this repo in a new directory.

        Note, this will copy all the files in the repo into a new local directory, but will not handle
        configuring the new directory as a git repo.
        """
        if await anyio.Path(new_repo_path).exists():
            if not exists_ok:
                raise FileExistsError(
                    f"New repo path '{new_repo_path} already exists. Set `exists_ok=True` if you are happy overwriting it, otherwise select new path."
                )
            await asyncio.to_thread(shutil.rmtree, new_repo_path)
        await asyncio.to_thread(
            shutil.copytree,
            self.base_path,
            new_repo_path,
            dirs_exist_ok=True,
            ignore=shutil.ignore_patterns(".git", ".gitsecret"),
        )
        return LocalGitRepo(new_repo_path)

    sync_copy_repo = sync(copy_repo)

    async def is_path_in_repo(self, file_path: Union[str, Path, anyio.Path]) -> bool:
        """Check whether a given file path is within this repo.

        FIXME: It doesn't seem entirely necessary to enumerate all of the files with a particular extension
               just to check if a single file (whose path we know) is in the repo.
        """
        if isinstance(file_path, (str, Path)):
            file_path = anyio.Path(file_path)
        extension = file_path.suffix
        return file_path in await self.get_all_files_by_extension(extension=extension)

    async def _get_file_path(self, file_path: Union[str, Path]) -> anyio.Path:
        path = anyio.Path(file_path)
        if not path.is_absolute():
            path = anyio.Path(self.base_path / path)
        assert await path.exists(), f"File {path} does not exist."
        return path

    async def safely_read_file_from_repo(self, file_path: Union[str, Path]) -> str:
        """Safely read file from repo."""
        path = await self._get_file_path(file_path)
        assert await self.is_path_in_repo(path), f"File {path} is not in repo."
        return await path.read_text()

    sync_safely_read_file_from_repo = sync(safely_read_file_from_repo)

    async def get_all_files_by_extension(self, extension: str = PYTHON_EXTENSION) -> Tuple[Path, ...]:
        """Get absolute path of all files in the repo with given extension."""
        paths: List[Path] = []
        async for path in anyio.Path(self.base_path).rglob(f"*{extension}"):
            paths.append(Path(path))
        return tuple(paths)


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class WritableLocalGitRepo(LocalGitRepo):
    """A Local Git Repo with support for modifying files and reseting to an initial state.

    Note, this does not handle creating a copy of an existing repo, or anything. Rather it adds some additional
    functionality for actually writing to files in the repo. For the creation of a separate copy of an existing
    repo which supports making changes without affecting the main repo see the `temp_writable_local_git_repo` context
    manager.

    It is also recommended the `build_from_repo` function when creating a WritableLocalGitRepo, as this will
    make sure that any untracked and uncommited changes are managed correctly.
    """

    initial_git_hash: str
    stash_git_hash: Optional[str]

    @classmethod
    async def build_from_repo(cls, repo: LocalGitRepo) -> "WritableLocalGitRepo":
        """Create a writable repo from an local repo."""
        init_hash = await repo.head_hash()
        if await is_repo_dirty(repo):
            await repo.run_git(("add", "."))
            stash_hash = await make_commit(repo, "stashing uncommited and untracked changes")
        else:
            stash_hash = None

        return cls(base_path=repo.base_path, initial_git_hash=init_hash, stash_git_hash=stash_hash)

    async def _setup(self) -> None:
        init_hash = await self.head_hash()
        expected_hash = self.stash_git_hash or self.initial_git_hash
        assert init_hash == expected_hash, "git repo is not currently at expected commit"
        assert await self.is_git_repo(), f"{self.base_path} is not a git repo"
        await self.assert_clean()

    async def reset(self) -> None:
        """Reset the repo to the state it was in when this class was created."""
        await restore_all_staged_files(self)
        await restore_all_unstaged_changes(self)
        if self.stash_git_hash:
            # hard to reset to commit with stashed untracked and uncommited changes
            await self.run_git(("reset", "--hard", self.stash_git_hash))
            # soft reset to return to initial commit but keep the untracked and uncommited changes
            await self.run_git(("reset", "--soft", self.initial_git_hash))
            # unstage untracked and uncommited changes
            await restore_all_staged_files(self)
        else:
            await self.run_git(("reset", "--hard", self.initial_git_hash))
            await self.run_git(("clean", "-f"))
            await self.run_git(("checkout", self.initial_git_hash))

        current_git_hash = await self.head_hash()
        assert (
            current_git_hash == self.initial_git_hash
        ), f"base branch changed, current git hash ({current_git_hash}) != initial git hash ({self.initial_git_hash})"
        await self.assert_clean()

    async def apply_change_to_file(self, file_path: Union[str, Path], new_contents: str) -> None:
        """Apply change to a single file."""
        path = await self._get_file_path(file_path)
        assert await self.is_path_in_repo(path), f"File {path} is not in repo."
        await path.write_text(new_contents)
        await git_add(self, str(path))

    async def __aenter__(self) -> "WritableLocalGitRepo":
        await self._setup()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.reset()


async def copy_files_from_one_repo_to_another(
    src_repo_path: Path, dst_repo_path: Path, relative_file_paths: Sequence[Union[str, Path]]
) -> None:
    """Copies files from src to dst repo using the relative file paths."""
    for relative_path in relative_file_paths:
        src_file_path = src_repo_path / relative_path
        dst_file_path = anyio.Path(dst_repo_path / relative_path)
        # make sure necessary directories exist in destination
        await dst_file_path.parent.mkdir(parents=True, exist_ok=True)
        await asyncio.to_thread(shutil.copy2, src_file_path, dst_file_path)


def get_repo_url_from_folder(repo_path: Path) -> str:
    try:
        repo_url = subprocess.check_output(
            ["git", "remote", "get-url", "origin"], cwd=repo_path, universal_newlines=True
        ).strip()
    except subprocess.CalledProcessError:
        raise
    else:
        if repo_url.startswith("git@"):
            # convert ssh url to https
            repo_url = repo_url.replace(":", "/")
            repo_url = f"https://{repo_url[4:]}"
        if "https://oauth2:" in repo_url:
            # remove the oauth2 prefix
            # repo_url is something like https://oauth2:{token}@gitlab.com/.../.git
            # change it to https://gitlab.com/.../.git
            # This will happen if repo was originallycloned using oauth2
            suffix = repo_url.split("@")[-1]
            repo_url = "https://" + suffix
        return repo_url


def get_repo_base_path() -> Path:
    working_directory = Path(__file__).parent
    try:
        return Path(
            _run_command_and_capture_output(["git", "rev-parse", "--show-toplevel"], cwd=working_directory).strip()
        )
    except subprocess.CalledProcessError as e:
        try:
            return working_directory.parents[1]
        except IndexError:
            raise UnableToFindRepoBase() from e


def _run_command_and_capture_output(args: Sequence[str], cwd: Optional[Path] = None) -> str:
    arg_str = " ".join(shlex.quote(arg) for arg in args)
    print(f"Running command: {arg_str}", file=sys.stderr)
    with subprocess.Popen(args, text=True, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as proc:
        with StringIO() as output:
            _handle_output(proc, output, sys.stderr)
            if proc.wait() != 0:
                raise subprocess.CalledProcessError(proc.returncode, cmd=args, output=output.getvalue())
            return output.getvalue()


class UnableToFindRepoBase(Exception):
    """Raised when the base of the repository cannot be found."""


def _handle_output(process: subprocess.Popen[str], *files: TextIO) -> None:
    process_stdout = process.stdout
    assert process_stdout is not None
    while True:
        output = process_stdout.read(1)
        if output:
            for f in files:
                f.write(output)
        elif process.poll() is not None:
            break


def get_diff_without_index(diff: str) -> str:
    new_lines = []
    for line in diff.splitlines():
        if line.startswith("index "):
            # We replace index lines with "index 0000000..0000000 100644" because:
            # - `0000000..0000000` ensures no real object hashes are referenced, making the diff neutral.
            # - `100644` is the standard file mode for non-executable files in git diffs, ensuring compatibility.
            # - This keeps the diff format valid while removing specific index information.
            new_lines.append("index 0000000..0000000 100644")
        else:
            new_lines.append(line)
    return "\n".join(new_lines).strip()


def is_diffs_without_index_equal(diff_1: str, diff_2: str) -> bool:
    return get_diff_without_index(diff_1) == get_diff_without_index(diff_2)


# Copy-pasted from imbue to avoid moving the whole hammers machinery over to imbue-core.
async def get_lines_from_process(shell_command: str, is_exit_code_validated: bool = True, **kwargs) -> List[str]:
    p = await asyncio.create_subprocess_shell(shell_command, stdin=PIPE, stdout=PIPE, stderr=STDOUT, **kwargs)
    lines = [x.decode("UTF-8") for x in (await p.communicate())[0].splitlines()]
    if is_exit_code_validated:
        joined_lines = "\n".join(lines)
        assert (
            p.returncode == 0
        ), f"command failed: {shell_command}\nwith output:\n{joined_lines} with exit code {p.returncode}"
    return lines



from enum import StrEnum
from pathlib import Path

IMBUE_TOML_PATH = Path("~/.imbue/config.toml").expanduser()

DEFAULT_PROJECT_SECRET_PATH = Path(".env")

LOW_PRIORITY_LEVEL = 37
MEDIUM_PRIORITY_LEVEL = 38
HIGH_PRIORITY_LEVEL = 39


class ExceptionPriority(StrEnum):
    # for issues that will result in the app crashing
    HIGH_PRIORITY = "HIGH_PRIORITY"
    # for issues that will cause major functionality to stop working
    MEDIUM_PRIORITY = "MEDIUM_PRIORITY"
    # everything else -- e.g. exception sites that are typically retriable, but may catch something unrecoverable
    LOW_PRIORITY = "LOW_PRIORITY"
